from collections.abc import Sequence
from datetime import datetime
from typing import AsyncGenerator, Optional
from uuid import UUID

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID, PageInfo

from ai.backend.common.data.model_deployment.types import ActivenessStatus as CommonActivenessStatus
from ai.backend.common.data.model_deployment.types import LivenessStatus as CommonLivenessStatus
from ai.backend.common.data.model_deployment.types import ReadinessStatus as CommonReadinessStatus
from ai.backend.common.exception import ModelDeploymentUnavailable
from ai.backend.manager.api.gql.base import (
    JSONString,
    OrderDirection,
    build_page_info,
    build_pagination_options,
    resolve_global_id,
    to_global_id,
)
from ai.backend.manager.api.gql.session import Session
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.data.deployment.types import ModelReplicaData, ReplicaOrderField
from ai.backend.manager.models.gql_models.session import ComputeSessionNode
from ai.backend.manager.repositories.deployment.types.types import (
    ActivenessStatusFilter as RepoActivenessStatus,
)
from ai.backend.manager.repositories.deployment.types.types import (
    ActivenessStatusFilterType,
    LivenessStatusFilterType,
    ModelReplicaFilterOptions,
    ModelReplicaOrderingOptions,
    ReadinessStatusFilterType,
)
from ai.backend.manager.repositories.deployment.types.types import (
    LivenessStatusFilter as RepoLivenessStatusFilter,
)
from ai.backend.manager.repositories.deployment.types.types import (
    ReadinessStatusFilter as RepoReadinessStatusFilter,
)
from ai.backend.manager.services.deployment.actions.batch_load_replicas_by_revision_ids import (
    BatchLoadReplicasByRevisionIdsAction,
)
from ai.backend.manager.services.deployment.actions.list_replicas import ListReplicasAction
from ai.backend.manager.types import PaginationOptions

from .model_revision import (
    ModelRevision,
)

ReadinessStatus = strawberry.enum(
    CommonReadinessStatus,
    name="ReadinessStatus",
    description="Added in 25.16.0. This enum represents the readiness status of a replica, indicating whether the deployment has been checked and its health state.",
)

LivenessStatus = strawberry.enum(
    CommonLivenessStatus,
    name="LivenessStatus",
    description="Added in 25.16.0. This enum represents the liveness status of a replica, indicating whether the deployment is currently running and able to serve requests.",
)

ActivenessStatus = strawberry.enum(
    CommonActivenessStatus,
    name="ActivenessStatus",
    description="Added in 25.16.0. This enum represents the activeness status of a replica, indicating whether the deployment is currently active and able to serve requests.",
)


@strawberry.input(description="Added in 25.16.0")
class ReadinessStatusFilter:
    in_: Optional[list[ReadinessStatus]] = strawberry.field(name="in", default=None)
    equals: Optional[ReadinessStatus] = None


@strawberry.input(description="Added in 25.16.0")
class LivenessStatusFilter:
    in_: Optional[list[LivenessStatus]] = strawberry.field(name="in", default=None)
    equals: Optional[LivenessStatus] = None


@strawberry.input(description="Added in 25.16.0")
class ActivenessStatusFilter:
    in_: Optional[list[ActivenessStatus]] = strawberry.field(name="in", default=None)
    equals: Optional[ActivenessStatus] = None


@strawberry.input(description="Added in 25.16.0")
class ReplicaFilter:
    readiness_status: Optional[ReadinessStatusFilter] = None
    liveness_status: Optional[LivenessStatusFilter] = None
    activeness_status: Optional[ActivenessStatusFilter] = None
    id: Optional[ID] = None
    ids_in: strawberry.Private[Optional[Sequence[UUID]]] = None

    AND: Optional[list["ReplicaFilter"]] = None
    OR: Optional[list["ReplicaFilter"]] = None
    NOT: Optional[list["ReplicaFilter"]] = None

    def to_repo_filter(self) -> ModelReplicaFilterOptions:
        repo_filter = ModelReplicaFilterOptions()

        if self.readiness_status:
            if self.readiness_status.in_:
                repo_filter.readiness_status_filter = RepoReadinessStatusFilter(
                    type=ReadinessStatusFilterType.IN,
                    values=[ReadinessStatus(status) for status in self.readiness_status.in_],
                )
            elif self.readiness_status.equals:
                repo_filter.readiness_status_filter = RepoReadinessStatusFilter(
                    type=ReadinessStatusFilterType.EQUALS,
                    values=[ReadinessStatus(self.readiness_status.equals)],
                )
        if self.liveness_status:
            if self.liveness_status.in_:
                repo_filter.liveness_status_filter = RepoLivenessStatusFilter(
                    type=LivenessStatusFilterType.IN,
                    values=[LivenessStatus(status) for status in self.liveness_status.in_],
                )
            elif self.liveness_status.equals:
                repo_filter.liveness_status_filter = RepoLivenessStatusFilter(
                    type=LivenessStatusFilterType.EQUALS,
                    values=[LivenessStatus(self.liveness_status.equals)],
                )
        if self.activeness_status:
            if self.activeness_status.in_:
                repo_filter.activeness_status_filter = RepoActivenessStatus(
                    type=ActivenessStatusFilterType.IN,
                    values=[ActivenessStatus(status) for status in self.activeness_status.in_],
                )
            elif self.activeness_status.equals:
                repo_filter.activeness_status_filter = RepoActivenessStatus(
                    type=ActivenessStatusFilterType.EQUALS,
                    values=[ActivenessStatus(self.activeness_status.equals)],
                )

        if self.id:
            repo_filter.id = UUID(self.id)
        if self.ids_in:
            repo_filter.ids_in = list(self.ids_in)

        # Handle logical operations
        if self.AND:
            repo_filter.AND = [f.to_repo_filter() for f in self.AND]
        if self.OR:
            repo_filter.OR = [f.to_repo_filter() for f in self.OR]
        if self.NOT:
            repo_filter.NOT = [f.to_repo_filter() for f in self.NOT]

        return repo_filter


@strawberry.input(description="Added in 25.16.0")
class ReplicaOrderBy:
    field: ReplicaOrderField
    direction: OrderDirection = OrderDirection.DESC


@strawberry.type(description="Added in 25.16.0")
class ModelReplica(Node):
    id: NodeID
    _session_id: strawberry.Private[UUID]
    _revision_id: strawberry.Private[UUID]
    readiness_status: ReadinessStatus = strawberry.field(
        description="This represents whether the replica has been checked and its health state.",
    )
    liveness_status: LivenessStatus = strawberry.field(
        description="This represents whether the replica is currently running and able to serve requests.",
    )
    activeness_status: ActivenessStatus = strawberry.field(
        description="This represents whether the replica is currently active and able to serve requests.",
    )
    weight: int
    detail: JSONString = strawberry.field(
        description="Detailed information about the routing node. It can include both error messages and success messages."
    )
    created_at: datetime
    live_stat: JSONString = strawberry.field(
        description='live statistics of the routing node. e.g. "live_stat": "{\\"cpu_util\\": {\\"current\\": \\"7.472\\", \\"capacity\\": \\"1000\\", \\"pct\\": \\"0.75\\", \\"unit_hint\\": \\"percent\\"}}"'
    )

    @strawberry.field(
        description="The session ID associated with the replica. This can be null right after replica creation."
    )
    async def session(self, info: Info[StrawberryGQLContext]) -> "Session":
        session_global_id = to_global_id(
            ComputeSessionNode, self._session_id, is_target_graphene_object=True
        )
        return Session(id=ID(session_global_id))

    @strawberry.field
    async def revision(self, info: Info[StrawberryGQLContext]) -> ModelRevision:
        """Resolve revision using dataloader."""
        revision_loader = info.context.dataloader_registry.get_loader(
            ModelRevision.batch_load_by_ids, info.context
        )
        revision: list[ModelRevision] = await revision_loader.load(self._revision_id)
        return revision[0]

    @classmethod
    def from_dataclass(cls, data: ModelReplicaData) -> "ModelReplica":
        return cls(
            id=ID(str(data.id)),
            _revision_id=data.revision_id,
            _session_id=data.session_id,
            readiness_status=ReadinessStatus(data.readiness_status),
            liveness_status=LivenessStatus(data.liveness_status),
            activeness_status=ActivenessStatus(data.activeness_status),
            weight=data.weight,
            detail=JSONString.serialize(data.detail),
            created_at=data.created_at,
            live_stat=JSONString.serialize(data.live_stat),
        )

    @classmethod
    async def batch_load_by_revision_ids(
        cls, ctx: StrawberryGQLContext, revision_ids: Sequence[UUID]
    ) -> list[list["ModelReplica"]]:
        """Batch load replicas by their revision IDs."""
        processor = ctx.processors.deployment
        if processor is None:
            raise ModelDeploymentUnavailable(
                "Model Deployment feature is unavailable. Please contact support."
            )

        action_result = await processor.batch_load_replicas_by_revision_ids.wait_for_complete(
            BatchLoadReplicasByRevisionIdsAction(revision_ids=list(revision_ids))
        )
        replicas_map = action_result.data

        result = []
        for revision_id in revision_ids:
            replica_data_list = replicas_map.get(revision_id, [])
            replica_objects = [
                cls.from_dataclass(replica_data) for replica_data in replica_data_list
            ]
            result.append(replica_objects)
        return result


ModelReplicaEdge = Edge[ModelReplica]


@strawberry.type(description="Added in 25.16.0")
class ModelReplicaConnection(Connection[ModelReplica]):
    count: int

    def __init__(self, *args, count: int, **kwargs):
        super().__init__(*args, **kwargs)
        self.count = count

    @classmethod
    def from_dataclass(cls, replicas_data: list[ModelReplicaData]) -> "ModelReplicaConnection":
        nodes = [ModelReplica.from_dataclass(data) for data in replicas_data]
        edges = [ModelReplicaEdge(node=node, cursor=str(node.id)) for node in nodes]

        page_info = PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        )

        return cls(count=len(nodes), edges=edges, page_info=page_info)


@strawberry.type(description="Added in 25.16.0")
class ReplicaStatusChangedPayload:
    replica: ModelReplica


@strawberry.field(description="Added in 25.16.0")
async def replica(id: ID, info: Info[StrawberryGQLContext]) -> Optional[ModelReplica]:
    """Get a specific replica by ID."""
    _, replica_id = resolve_global_id(id)
    replica_loader = info.context.dataloader_registry.get_loader(
        ModelReplica.batch_load_by_revision_ids, info.context
    )
    replicas: list[ModelReplica] = await replica_loader.load(UUID(replica_id))
    return replicas[0]


def _convert_gql_replica_ordering_to_repo_ordering(
    order_by: Optional[list[ReplicaOrderBy]],
) -> ModelReplicaOrderingOptions:
    if not order_by:
        return ModelReplicaOrderingOptions()

    repo_order_by = []
    for order in order_by:
        desc = order.direction == OrderDirection.DESC
        repo_order_by.append((order.field, desc))

    return ModelReplicaOrderingOptions(order_by=repo_order_by)


async def resolve_replicas(
    info: Info[StrawberryGQLContext],
    filter: Optional[ReplicaFilter] = None,
    order_by: Optional[list[ReplicaOrderBy]] = None,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> ModelReplicaConnection:
    repo_filter = None
    if filter:
        repo_filter = filter.to_repo_filter()

    repo_ordering = _convert_gql_replica_ordering_to_repo_ordering(order_by)

    pagination_options = build_pagination_options(
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )

    processor = info.context.processors.deployment
    if processor is None:
        raise ModelDeploymentUnavailable(
            "Model Deployment feature is unavailable. Please contact support."
        )

    action_result = await processor.list_replicas.wait_for_complete(
        ListReplicasAction(
            pagination=PaginationOptions(),
            ordering=repo_ordering,
            filters=repo_filter,
        )
    )
    edges = []
    for replica_data in action_result.data:
        node = ModelReplica.from_dataclass(replica_data)
        edge = ModelReplicaEdge(node=node, cursor=str(node.id))
        edges.append(edge)

    page_info = build_page_info(
        edges=edges, total_count=action_result.total_count, pagination_options=pagination_options
    )

    return ModelReplicaConnection(
        count=action_result.total_count,
        edges=edges,
        page_info=page_info.to_strawberry_page_info(),
    )


@strawberry.field(description="Added in 25.16.0")
async def replicas(
    info: Info[StrawberryGQLContext],
    filter: Optional[ReplicaFilter] = None,
    order_by: Optional[list[ReplicaOrderBy]] = None,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> ModelReplicaConnection:
    return await resolve_replicas(
        info=info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.subscription(description="Added in 25.16.0")
async def replica_status_changed(
    revision_id: ID,
) -> AsyncGenerator[ReplicaStatusChangedPayload, None]:
    """Subscribe to replica status changes."""
    if False:  # Replace with actual subscription logic
        yield ReplicaStatusChangedPayload(replica=replica)



import logging
from typing import Callable

from ai.backend.common.data.artifact.types import ArtifactRegistryType
from ai.backend.common.events.event_types.artifact_registry.anycast import (
    DoScanReservoirRegistryEvent,
)
from ai.backend.common.types import (
    AgentId,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.clients.storage_proxy.session_manager import StorageSessionManager
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.artifact.types import (
    ArtifactType,
)
from ai.backend.manager.data.object_storage.types import ObjectStorageData
from ai.backend.manager.data.vfs_storage.types import VFSStorageData
from ai.backend.manager.errors.artifact_registry import ReservoirConnectionError
from ai.backend.manager.repositories.artifact.repository import ArtifactRepository
from ai.backend.manager.repositories.artifact_registry.repository import ArtifactRegistryRepository
from ai.backend.manager.repositories.object_storage.repository import ObjectStorageRepository
from ai.backend.manager.repositories.reservoir_registry.repository import (
    ReservoirRegistryRepository,
)
from ai.backend.manager.repositories.vfs_storage.repository import VFSStorageRepository
from ai.backend.manager.services.artifact.actions.scan import ScanArtifactsAction
from ai.backend.manager.services.processors import Processors

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ArtifactRegistryEventHandler:
    _processors_factory: Callable[[], Processors]
    _artifact_repository: ArtifactRepository
    _artifact_registry_repository: ArtifactRegistryRepository
    _reservoir_registry_repository: ReservoirRegistryRepository
    _object_storage_repository: ObjectStorageRepository
    _vfs_storage_repository: VFSStorageRepository
    _storage_manager: StorageSessionManager
    _config_provider: ManagerConfigProvider

    def __init__(
        self,
        processors_factory: Callable[[], Processors],
        artifact_repository: ArtifactRepository,
        artifact_registry_repository: ArtifactRegistryRepository,
        reservoir_registry_repository: ReservoirRegistryRepository,
        object_storage_repository: ObjectStorageRepository,
        vfs_storage_repository: VFSStorageRepository,
        storage_manager: StorageSessionManager,
        config_provider: ManagerConfigProvider,
    ) -> None:
        self._processors_factory = processors_factory
        self._artifact_repository = artifact_repository
        self._artifact_registry_repository = artifact_registry_repository
        self._reservoir_registry_repository = reservoir_registry_repository
        self._object_storage_repository = object_storage_repository
        self._vfs_storage_repository = vfs_storage_repository
        self._storage_manager = storage_manager
        self._config_provider = config_provider

    async def handle_artifact_registry_scan(
        self, context: None, source: AgentId, event: DoScanReservoirRegistryEvent
    ) -> None:
        processors = self._processors_factory()
        registries = await self._artifact_registry_repository.list_artifact_registry_data()

        for registry in registries:
            if registry.type != ArtifactRegistryType.RESERVOIR:
                continue

            try:
                await processors.artifact.scan.wait_for_complete(
                    ScanArtifactsAction(
                        registry_id=registry.registry_id,
                        # TODO: Support other artifact types in the future
                        artifact_type=ArtifactType.MODEL,
                        # Fetch all artifacts without limit
                        limit=None,
                        order=None,
                        search=None,
                    )
                )
                log.info("Completed scanning reservoir registry: {}.", registry.registry_id)
            except ReservoirConnectionError:
                log.warning(
                    "Failed to scan reservoir registry: {}.",
                    registry.registry_id,
                )

    async def _resolve_storage_data(self, storage_name: str) -> ObjectStorageData | VFSStorageData:
        try:
            return await self._object_storage_repository.get_by_name(storage_name)
        except Exception:
            return await self._vfs_storage_repository.get_by_name(storage_name)



"""Database source for resource preset repository operations."""

from __future__ import annotations

import logging
from collections import defaultdict
from typing import Mapping, Optional
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.common.exception import ResourcePresetConflict
from ai.backend.common.types import (
    AccessKey,
    AgentId,
    DefaultForUnspecified,
    ResourceSlot,
    SlotName,
    SlotTypes,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.resource_preset.types import ResourcePresetData
from ai.backend.manager.errors.resource import (
    DomainNotFound,
    InvalidPresetQuery,
    ProjectNotFound,
    ResourcePresetNotFound,
    ScalingGroupNotFound,
)
from ai.backend.manager.models import (
    AgentRow,
    KernelRow,
    SessionRow,
    association_groups_users,
    domains,
    groups,
    query_allowed_sgroups,
)
from ai.backend.manager.models.resource_preset import ResourcePresetRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.services.resource_preset.types import (
    ResourcePresetCreator,
    ResourcePresetModifier,
)

from .types import (
    AccessKeyFilter,
    CheckPresetsDBData,
    DomainNameFilter,
    GroupIdFilter,
    KeypairResourceData,
    PerScalingGroupResourceData,
    PresetAllocatabilityData,
    ResourceOccupancyFilter,
    ResourceUsageData,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ResourcePresetDBSource:
    """Database source for resource preset operations."""

    _db: ExtendedAsyncSAEngine

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
    ) -> None:
        self._db = db

    async def create_preset(self, creator: ResourcePresetCreator) -> ResourcePresetData:
        """
        Creates a new resource preset.
        Raises ResourcePresetConflict if a preset with the same name and scaling group already exists.
        """
        async with self._db.begin_session() as session:
            preset_row = await ResourcePresetRow.create(creator, db_session=session)
            if preset_row is None:
                raise ResourcePresetConflict(
                    f"Duplicate resource preset name (name:{creator.name}, scaling_group:{creator.scaling_group_name})"
                )
            data = preset_row.to_dataclass()
        return data

    async def get_preset_by_id(self, preset_id: UUID) -> ResourcePresetData:
        """
        Gets a resource preset by ID.
        Raises ResourcePresetNotFound if the preset doesn't exist.
        """
        async with self._db.begin_readonly_session() as session:
            preset_row = await self._get_preset_by_id(session, preset_id)
            if preset_row is None:
                raise ResourcePresetNotFound()
            data = preset_row.to_dataclass()
        return data

    async def get_preset_by_name(self, name: str) -> ResourcePresetData:
        """
        Gets a resource preset by name.
        Raises ResourcePresetNotFound if the preset doesn't exist.
        """
        async with self._db.begin_readonly_session() as session:
            preset_row = await self._get_preset_by_name(session, name)
            if preset_row is None:
                raise ResourcePresetNotFound()
            data = preset_row.to_dataclass()
        return data

    async def get_preset_by_id_or_name(
        self, preset_id: Optional[UUID], name: Optional[str]
    ) -> ResourcePresetData:
        """
        Gets a resource preset by ID or name.
        ID takes precedence if both are provided.
        Raises ResourcePresetNotFound if the preset doesn't exist.
        """
        async with self._db.begin_readonly_session() as session:
            preset_row = await self._get_preset_by_id_or_name(session, preset_id, name)
            data = preset_row.to_dataclass()
        return data

    async def _get_preset_by_id_or_name(
        self, db_sess: SASession, preset_id: Optional[UUID], name: Optional[str]
    ) -> ResourcePresetRow:
        if preset_id is not None:
            preset_row = await self._get_preset_by_id(db_sess, preset_id)
        elif name is not None:
            preset_row = await self._get_preset_by_name(db_sess, name)
        else:
            raise InvalidPresetQuery("Either preset_id or name must be provided")

        if preset_row is None:
            raise ResourcePresetNotFound()
        return preset_row

    async def modify_preset(
        self, preset_id: Optional[UUID], name: Optional[str], modifier: ResourcePresetModifier
    ) -> ResourcePresetData:
        """
        Modifies an existing resource preset.
        Raises ResourcePresetNotFound if the preset doesn't exist.
        """
        async with self._db.begin_session() as session:
            preset_row = await self._get_preset_by_id_or_name(session, preset_id, name)
            to_update = modifier.fields_to_update()
            for key, value in to_update.items():
                setattr(preset_row, key, value)
            await session.flush()
            data = preset_row.to_dataclass()
        return data

    async def delete_preset(
        self, preset_id: Optional[UUID], name: Optional[str]
    ) -> ResourcePresetData:
        """
        Deletes a resource preset.
        Returns the deleted preset data.
        Raises ResourcePresetNotFound if the preset doesn't exist.
        """
        async with self._db.begin_session() as session:
            preset_row = await self._get_preset_by_id_or_name(session, preset_id, name)
            data = preset_row.to_dataclass()
            await session.delete(preset_row)
        return data

    async def list_presets(
        self, scaling_group_name: Optional[str] = None
    ) -> list[ResourcePresetData]:
        """
        Lists all resource presets.
        If scaling_group_name is provided, returns presets for that scaling group and global presets.
        """
        async with self._db.begin_readonly_session() as session:
            query = sa.select(ResourcePresetRow)
            if scaling_group_name is not None:
                query = query.where(
                    sa.or_(
                        ResourcePresetRow.scaling_group_name.is_(None),
                        ResourcePresetRow.scaling_group_name == scaling_group_name,
                    )
                )
            else:
                query = query.where(ResourcePresetRow.scaling_group_name.is_(None))

            presets = []
            async for row in await session.stream_scalars(query):
                presets.append(row.to_dataclass())

        return presets

    async def check_presets_data(
        self,
        access_key: AccessKey,
        user_id: UUID,
        group_name: str,
        domain_name: str,
        resource_policy: Mapping[str, str],
        known_slot_types: Mapping[SlotName, SlotTypes],
        scaling_group: Optional[str] = None,
    ) -> CheckPresetsDBData:
        """
        Fetch all data needed for checking presets from database.
        This includes resource limits, occupancy, and preset allocatability.
        """
        async with self._db.begin_readonly_session() as conn:
            # Fetch all database data at once
            db_data = await self._fetch_all_check_presets_data(
                conn,
                access_key,
                user_id,
                group_name,
                domain_name,
                resource_policy,
                known_slot_types,
                scaling_group,
            )

        return db_data

    async def _get_group_info(
        self,
        db_sess: SASession,
        user_id: UUID,
        group_name: str,
        domain_name: str,
    ) -> tuple[UUID, Mapping[str, str]]:
        """
        Get group ID and resource slots for a user's group.

        :param db_sess: Database session
        :param user_id: User ID
        :param group_name: Group name
        :param domain_name: Domain name
        :return: Tuple of (group_id, total_resource_slots)
        :raises ProjectNotFound: If the group does not exist or the user is not a member
        """
        j = sa.join(
            groups,
            association_groups_users,
            association_groups_users.c.group_id == groups.c.id,
        )
        query = (
            sa.select([groups.c.id, groups.c.total_resource_slots])
            .select_from(j)
            .where(
                (association_groups_users.c.user_id == user_id)
                & (groups.c.name == group_name)
                & (groups.c.domain_name == domain_name),
            )
        )
        result = await db_sess.execute(query)
        row = result.first()
        if row is None:
            raise ProjectNotFound(f"Project not found (name: {group_name})")

        return row["id"], row["total_resource_slots"]

    async def _get_domain_resource_slots(
        self,
        db_sess: SASession,
        domain_name: str,
        known_slot_types: Mapping[SlotName, SlotTypes],
    ) -> ResourceUsageData:
        """
        Get domain resource slots.

        :param db_sess: Database session
        :param domain_name: Domain name
        :return: ResourceUsageData with domain resource slots
        """
        query = sa.select([domains.c.total_resource_slots]).where(domains.c.name == domain_name)
        result = await db_sess.execute(query)
        domain_resource_slots = result.first()[0]
        if domain_resource_slots is None:
            raise DomainNotFound(f"Domain not found (name: {domain_name})")
        domain_resource_policy = {
            "total_resource_slots": domain_resource_slots,
            "default_for_unspecified": DefaultForUnspecified.UNLIMITED,
        }
        limits = ResourceSlot.from_policy(domain_resource_policy, known_slot_types)
        occupied = await self._get_resource_occupancy(
            db_sess, known_slot_types, filters=[DomainNameFilter(domain_name)]
        )
        remaining = limits - occupied
        return ResourceUsageData(
            limits=limits,
            occupied=occupied,
            remaining=remaining,
        )

    async def _get_user_session_occupancy_per_sgroup(
        self,
        db_sess: SASession,
        user_id: UUID,
        sgroup_names: list[str],
        known_slot_types: Mapping[SlotName, SlotTypes],
    ) -> dict[str, ResourceSlot]:
        """
        Get user's session resource occupancy per scaling group.

        :param db_sess: Database session
        :param user_id: User ID
        :param sgroup_names: List of scaling group names
        :param known_slot_types: Known slot types for initialization
        :return: Dictionary of scaling group name to occupied resources
        """
        per_sgroup_occupancy = {
            sgname: ResourceSlot.from_known_slots(known_slot_types) for sgname in sgroup_names
        }

        j = sa.join(KernelRow, SessionRow, KernelRow.session_id == SessionRow.id)
        query = (
            sa.select([KernelRow.occupied_slots, SessionRow.scaling_group_name])
            .select_from(j)
            .where(
                (KernelRow.user_uuid == user_id)
                & (KernelRow.status.in_(KernelStatus.resource_occupied_statuses()))
                & (SessionRow.scaling_group_name.in_(sgroup_names)),
            )
        )
        async for row in await db_sess.stream(query):
            if row["occupied_slots"]:
                per_sgroup_occupancy[row["scaling_group_name"]] += row["occupied_slots"]

        return per_sgroup_occupancy

    async def _get_agent_available_resources(
        self,
        db_sess: SASession,
        sgroup_names: list[str],
        known_slot_types: Mapping[SlotName, SlotTypes],
    ) -> tuple[dict[str, ResourceSlot], list[ResourceSlot]]:
        """
        Get available resources from agents in given scaling groups.

        Calculate actual occupied slots by aggregating from kernels with
        resource_occupied_statuses (RUNNING, TERMINATING) instead of using
        the cached AgentRow.occupied_slots value.

        Uses two efficient queries: one for agents, and one for filtered kernels
        only with resource_occupied_statuses to minimize data loading.

        :param db_sess: Database session
        :param sgroup_names: List of scaling group names
        :param known_slot_types: Known slot types for initialization
        :return: Tuple of (per_sgroup_remaining, agent_slots_list)
        """
        # Query 1: Get agents in the scaling groups
        agent_query = sa.select(AgentRow).where(
            sa.and_(
                AgentRow.scaling_group.in_(sgroup_names),
                AgentRow.available_slots.isnot(None),
                AgentRow.schedulable == sa.true(),
            )
        )

        agent_result = await db_sess.execute(agent_query)
        agent_rows = list(agent_result.scalars().all())

        if not agent_rows:
            # No agents found, return empty results
            return (
                {
                    sgname: ResourceSlot.from_known_slots(known_slot_types)
                    for sgname in sgroup_names
                },
                [],
            )

        # Query 2: Get only kernels with resource_occupied_statuses for these agents
        agent_ids = [agent.id for agent in agent_rows]
        kernel_query = sa.select(KernelRow.agent, KernelRow.occupied_slots).where(
            sa.and_(
                KernelRow.agent.in_(agent_ids),
                KernelRow.status.in_(KernelStatus.resource_occupied_statuses()),
            )
        )

        kernel_result = await db_sess.execute(kernel_query)

        # Aggregate occupied slots by agent
        agent_occupied: dict[AgentId, ResourceSlot] = defaultdict(
            lambda: ResourceSlot.from_known_slots(known_slot_types)
        )
        for row in kernel_result:
            if row.agent and row.occupied_slots:
                agent_occupied[row.agent] += row.occupied_slots

        # Calculate remaining resources per agent and per scaling group
        per_sgroup_remaining = {
            sgname: ResourceSlot.from_known_slots(known_slot_types) for sgname in sgroup_names
        }
        agent_slots = []

        for agent in agent_rows:
            actual_occupied = agent_occupied[agent.id]
            remaining = agent.available_slots - actual_occupied
            agent_slots.append(remaining)
            per_sgroup_remaining[agent.scaling_group] += remaining

        return per_sgroup_remaining, agent_slots

    async def _get_resource_occupancy(
        self,
        db_sess: SASession,
        known_slot_types: Mapping[SlotName, SlotTypes],
        filters: list[ResourceOccupancyFilter] | None = None,
    ) -> ResourceSlot:
        """
        Get resource occupancy with filters.

        :param db_sess: Database session
        :param known_slot_types: Known slot types for initialization
        :param filters: List of filter objects to apply
        :return: Total occupied resources
        """
        conditions = [KernelRow.status.in_(KernelStatus.resource_occupied_statuses())]

        if filters:
            for filter_obj in filters:
                conditions.append(filter_obj.get_condition())

        query = sa.select([KernelRow.occupied_slots]).where(sa.and_(*conditions))

        total = ResourceSlot.from_known_slots(known_slot_types)
        async for row in await db_sess.stream(query):
            if row[0]:  # occupied_slots might be null
                total += row[0]
        return total

    async def _get_keypair_resource_usage(
        self,
        db_sess: SASession,
        access_key: AccessKey,
        resource_policy: Mapping[str, str],
        known_slot_types: Mapping[SlotName, SlotTypes],
    ) -> ResourceUsageData:
        """Get keypair resource usage (limits, occupied, remaining)."""
        limits = ResourceSlot.from_policy(resource_policy, known_slot_types)
        occupied = await self._get_resource_occupancy(
            db_sess, known_slot_types, filters=[AccessKeyFilter(access_key)]
        )
        remaining = limits - occupied

        return ResourceUsageData(
            limits=limits,
            occupied=occupied,
            remaining=remaining,
        )

    async def _get_group_resource_usage(
        self,
        db_sess: SASession,
        group_id: UUID,
        group_resource_slots: Mapping[str, str],
        known_slot_types: Mapping[SlotName, SlotTypes],
    ) -> ResourceUsageData:
        """Get group resource usage (limits, occupied, remaining)."""
        group_resource_policy = {
            "total_resource_slots": group_resource_slots,
            "default_for_unspecified": DefaultForUnspecified.UNLIMITED,
        }
        limits = ResourceSlot.from_policy(group_resource_policy, known_slot_types)
        occupied = await self._get_resource_occupancy(
            db_sess, known_slot_types, filters=[GroupIdFilter(group_id)]
        )
        remaining = limits - occupied

        return ResourceUsageData(
            limits=limits,
            occupied=occupied,
            remaining=remaining,
        )

    async def _fetch_all_check_presets_data(
        self,
        conn: SASession,
        access_key: AccessKey,
        user_id: UUID,
        group_name: str,
        domain_name: str,
        resource_policy: Mapping[str, str],
        known_slot_types: Mapping[SlotName, SlotTypes],
        scaling_group: Optional[str] = None,
    ) -> CheckPresetsDBData:
        """
        Fetch all data needed for check_presets in a single method.
        """
        # Get keypair resource usage
        keypair_usage = await self._get_keypair_resource_usage(
            conn, access_key, resource_policy, known_slot_types
        )

        # Get group info and resource usage
        group_id, group_resource_slots = await self._get_group_info(
            conn, user_id, group_name, domain_name
        )
        group_usage = await self._get_group_resource_usage(
            conn, group_id, group_resource_slots, known_slot_types
        )

        # Get domain resource slots and usage
        domain_usage = await self._get_domain_resource_slots(conn, domain_name, known_slot_types)
        # Take minimum remaining resources across all scopes
        final_remaining = ResourceSlot.from_known_slots(known_slot_types)
        for slot in known_slot_types:
            final_remaining[slot] = min(
                keypair_usage.remaining[slot],
                group_usage.remaining[slot],
                domain_usage.remaining[slot],
            )

        # Get scaling groups
        sgroups = await query_allowed_sgroups(conn, domain_name, group_id, access_key)
        sgroup_names = [sg.name for sg in sgroups]
        if scaling_group is not None:
            if scaling_group not in sgroup_names:
                raise ScalingGroupNotFound(
                    f"Scaling group not found or not allowed (name: {scaling_group})"
                )
            sgroup_names = [scaling_group]

        # Get user's session occupancy per scaling group
        per_sgroup_occupancy = await self._get_user_session_occupancy_per_sgroup(
            conn, user_id, sgroup_names, known_slot_types
        )

        # Get agent available resources per scaling group
        per_sgroup_agent_remaining, agent_slots = await self._get_agent_available_resources(
            conn, sgroup_names, known_slot_types
        )

        # Build per scaling group data
        per_sgroup = {}
        empty_slot = ResourceSlot.from_known_slots(known_slot_types)
        for sgname in sgroup_names:
            per_sgroup[sgname] = PerScalingGroupResourceData(
                using=per_sgroup_occupancy.get(sgname, empty_slot),
                remaining=per_sgroup_agent_remaining.get(sgname, empty_slot),
            )

        # Calculate total scaling group remaining
        sgroup_remaining = ResourceSlot.from_known_slots(known_slot_types)
        for remaining in per_sgroup_agent_remaining.values():
            sgroup_remaining += remaining

        # Apply final remaining limits to per scaling group remaining
        for sgname, sg_data in per_sgroup.items():
            for slot in known_slot_types.keys():
                if slot in sg_data.remaining:
                    sg_data.remaining[slot] = min(final_remaining[slot], sg_data.remaining[slot])

        for slot in known_slot_types.keys():
            sgroup_remaining[slot] = min(final_remaining[slot], sgroup_remaining[slot])

        # Fetch resource presets
        preset_data_list = await self.list_presets(scaling_group)

        # Check preset allocatability
        presets = []
        for preset_data in preset_data_list:
            allocatable = False
            preset_slots = preset_data.resource_slots.normalize_slots(ignore_unknown=True)
            for agent_slot in agent_slots:
                if agent_slot >= preset_slots and final_remaining >= preset_slots:
                    allocatable = True
                    break

            presets.append(
                PresetAllocatabilityData(
                    preset=preset_data,
                    allocatable=allocatable,
                )
            )

        # Build KeypairResourceData with all usage information
        keypair_data = KeypairResourceData(
            limits=keypair_usage.limits,
            occupied=keypair_usage.occupied,
            remaining=final_remaining,  # Use the minimum remaining across all scopes
            group_limits=group_usage.limits,
            group_occupied=group_usage.occupied,
            group_remaining=group_usage.remaining,
            scaling_group_remaining=sgroup_remaining,
        )

        return CheckPresetsDBData(
            known_slot_types=known_slot_types,
            keypair_data=keypair_data,
            per_sgroup_data=per_sgroup,
            presets=presets,
        )

    async def _get_preset_by_id(
        self, session: SASession, preset_id: UUID
    ) -> Optional[ResourcePresetRow]:
        """
        Private method to get a preset by ID using an existing session.
        """
        return await session.scalar(
            sa.select(ResourcePresetRow).where(ResourcePresetRow.id == preset_id)
        )

    async def _get_preset_by_name(
        self, session: SASession, name: str
    ) -> Optional[ResourcePresetRow]:
        """
        Private method to get a preset by name using an existing session.
        """
        return await session.scalar(
            sa.select(ResourcePresetRow).where(ResourcePresetRow.name == name)
        )



from collections.abc import Mapping
from dataclasses import dataclass, field
from datetime import datetime
from functools import lru_cache
from typing import Any, Optional
from uuid import UUID

from ai.backend.common.data.vfolder.types import VFolderMountData
from ai.backend.common.types import (
    AccessKey,
    CIStrEnum,
    ClusterMode,
    ResourceSlot,
    SessionId,
    SessionResult,
    SessionTypes,
)
from ai.backend.manager.data.user.types import UserData
from ai.backend.manager.models.network import NetworkType


class SessionStatus(CIStrEnum):
    # values are only meaningful inside the manager
    PENDING = "PENDING"
    # ---
    SCHEDULED = "SCHEDULED"
    PREPARING = "PREPARING"
    # manager can set PENDING, SCHEDULED and PREPARING independently
    # ---
    PULLING = "PULLING"
    PREPARED = "PREPARED"
    CREATING = "CREATING"
    # ---
    RUNNING = "RUNNING"
    RESTARTING = "RESTARTING"
    RUNNING_DEGRADED = "RUNNING_DEGRADED"
    # ---
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"
    ERROR = "ERROR"
    CANCELLED = "CANCELLED"

    @classmethod
    def kernel_awaiting_statuses(cls) -> set["SessionStatus"]:
        return {
            cls.PREPARING,
            cls.PULLING,
            cls.CREATING,
            cls.TERMINATING,
        }

    @classmethod
    @lru_cache(maxsize=1)
    def resource_occupied_statuses(cls) -> frozenset["SessionStatus"]:
        return frozenset(
            status
            for status in cls
            if status
            not in (
                cls.PENDING,
                cls.TERMINATED,
                cls.CANCELLED,
            )
        )

    @classmethod
    @lru_cache(maxsize=1)
    def terminatable_statuses(cls) -> frozenset["SessionStatus"]:
        """Return statuses that can transition to TERMINATING."""
        return frozenset(
            status
            for status in cls
            if status
            not in (
                cls.PENDING,
                cls.TERMINATING,
                cls.TERMINATED,
                cls.CANCELLED,
                cls.ERROR,
            )
        )

    @classmethod
    @lru_cache(maxsize=1)
    def terminal_statuses(cls) -> frozenset["SessionStatus"]:
        return frozenset((
            cls.ERROR,
            cls.TERMINATED,
            cls.CANCELLED,
        ))

    @classmethod
    @lru_cache(maxsize=1)
    def retriable_statuses(cls) -> frozenset["SessionStatus"]:
        return frozenset(
            (
                status
                for status in cls
                if status
                not in (
                    cls.RUNNING,
                    cls.TERMINATING,
                    cls.TERMINATED,
                    cls.CANCELLED,
                )
            )
        )

    def is_terminal(self) -> bool:
        return self in self.terminal_statuses()


# TODO: Add proper types
@dataclass
class SessionData:
    id: UUID
    session_type: SessionTypes
    priority: int
    cluster_mode: ClusterMode
    cluster_size: int
    domain_name: str
    group_id: UUID
    user_uuid: UUID
    occupying_slots: Any  # TODO: ResourceSlot?
    requested_slots: Any
    use_host_network: bool
    created_at: datetime = field(compare=False)
    status: "SessionStatus"
    result: SessionResult
    num_queries: int
    creation_id: Optional[str]
    name: Optional[str]
    access_key: Optional[AccessKey]
    agent_ids: Optional[list[str]]
    images: Optional[list[str]]
    tag: Optional[str]
    vfolder_mounts: Optional[list[VFolderMountData]]
    environ: Optional[dict[str, Any]]
    bootstrap_script: Optional[str]
    target_sgroup_names: Optional[list[str]]
    timeout: Optional[int]
    batch_timeout: Optional[int]
    terminated_at: Optional[datetime] = field(compare=False)
    scaling_group_name: Optional[str]
    starts_at: Optional[datetime] = field(compare=False)
    status_info: Optional[str] = field(compare=False)
    status_data: Optional[dict[str, Any]] = field(compare=False)
    status_history: Optional[dict[str, Any]] = field(compare=False)
    callback_url: Optional[str]
    startup_command: Optional[str]
    last_stat: Optional[dict[str, Any]] = field(compare=False)
    network_type: Optional[NetworkType]
    network_id: Optional[str]
    owner: Optional[UserData] = field(compare=False)

    # Loaded from relationship
    service_ports: Optional[str]


@dataclass
class SessionIdentity:
    id: SessionId
    creation_id: str
    name: str
    session_type: SessionTypes
    priority: int


@dataclass
class SessionMetadata:
    name: str
    domain_name: str
    group_id: UUID
    user_uuid: UUID
    access_key: str
    session_type: SessionTypes
    priority: int
    created_at: Optional[datetime]
    tag: Optional[str]


@dataclass
class ResourceSpec:
    cluster_mode: str
    cluster_size: int
    occupying_slots: ResourceSlot
    requested_slots: ResourceSlot
    scaling_group_name: Optional[str]
    target_sgroup_names: Optional[list[str]]
    agent_ids: Optional[list[str]]


@dataclass
class ImageSpec:
    images: Optional[list[str]]
    tag: Optional[str]


@dataclass
class MountSpec:
    vfolder_mounts: Optional[list[dict[str, Any]]]


@dataclass
class SessionExecution:
    environ: Optional[dict[str, Any]]
    bootstrap_script: Optional[str]
    startup_command: Optional[str]
    use_host_network: bool
    callback_url: Optional[str]


@dataclass
class SessionLifecycle:
    status: SessionStatus
    result: SessionResult
    created_at: Optional[datetime]
    terminated_at: Optional[datetime]
    starts_at: Optional[datetime]
    status_changed: Optional[datetime]
    batch_timeout: Optional[int]
    status_info: Optional[str]
    status_data: Optional[Mapping[str, Any]]
    status_history: Optional[dict[str, Any]]


@dataclass
class SessionMetrics:
    num_queries: int
    last_stat: Optional[dict[str, Any]]


@dataclass
class SessionNetwork:
    network_type: Optional[NetworkType]
    network_id: Optional[str]


@dataclass
class SessionInfo:
    identity: SessionIdentity
    metadata: SessionMetadata
    resource: ResourceSpec
    image: ImageSpec
    mounts: MountSpec
    execution: SessionExecution
    lifecycle: SessionLifecycle
    metrics: SessionMetrics
    network: SessionNetwork



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.artifact_registries.types import ArtifactRegistryData
from ai.backend.manager.services.artifact_registry.actions.base import ArtifactRegistryAction


@dataclass
class GetArtifactRegistryMetasAction(ArtifactRegistryAction):
    registry_ids: list[uuid.UUID]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "get_meta_multi"


@dataclass
class GetArtifactRegistryMetasActionResult(BaseActionResult):
    result: list[ArtifactRegistryData]

    @override
    def entity_id(self) -> Optional[str]:
        return None



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.notification import NotificationRuleData
from ai.backend.manager.repositories.base import Querier

from .base import NotificationAction


@dataclass
class SearchRulesAction(NotificationAction):
    """Action to search notification rules."""

    querier: Querier

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "search_rules"

    @override
    def entity_id(self) -> Optional[str]:
        return None


@dataclass
class SearchRulesActionResult(BaseActionResult):
    """Result of searching notification rules."""

    rules: list[NotificationRuleData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool

    @override
    def entity_id(self) -> Optional[str]:
        return None



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.domain.admin_repository import AdminDomainRepository
from ai.backend.manager.repositories.domain.repository import DomainRepository
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class DomainRepositories:
    repository: DomainRepository
    admin_repository: AdminDomainRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = DomainRepository(args.db)
        admin_repository = AdminDomainRepository(args.db)

        return cls(
            repository=repository,
            admin_repository=admin_repository,
        )



import uuid
from dataclasses import dataclass

from ..user import UserRole
from ..utils import ExtendedAsyncSAEngine


@dataclass
class ClientContext:
    db: ExtendedAsyncSAEngine

    domain_name: str
    user_id: uuid.UUID
    user_role: UserRole



"""Validator for reserved batch sessions."""

from datetime import datetime

from dateutil.tz import tzutc

from ai.backend.common.types import SessionTypes
from ai.backend.manager.sokovan.scheduler.types import SessionWorkload, SystemSnapshot

from .exceptions import SchedulingValidationError
from .validator import ValidatorRule


class ReservedBatchSessionValidator(ValidatorRule):
    """
    Check if a batch-type session should not be started for a certain amount of time.
    This corresponds to check_reserved_batch_session predicate.
    """

    def name(self) -> str:
        """Return the validator name for predicates."""
        return "ReservedBatchSessionValidator"

    def success_message(self) -> str:
        """Return a message describing successful validation."""
        return "Batch session has reached its scheduled start time"

    def validate(self, snapshot: SystemSnapshot, workload: SessionWorkload) -> None:
        # Check if this is a batch session with a scheduled start time
        if workload.session_type == SessionTypes.BATCH and workload.starts_at is not None:
            # Check if the current time is before the scheduled start time
            if datetime.now(tzutc()) < workload.starts_at:
                raise SchedulingValidationError("Before start time")



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.image.types import (
    ImageIdentifier,
    ImageStatus,
    ImageWithAgentInstallStatus,
)
from ai.backend.manager.data.user.types import UserRole
from ai.backend.manager.services.image.actions.base import ImageAction


@dataclass
class GetImageByIdentifierAction(ImageAction):
    image_identifier: ImageIdentifier
    user_role: UserRole
    image_status: Optional[list[ImageStatus]]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "get_by_identifier"


@dataclass
class GetImageByIdentifierActionResult(BaseActionResult):
    image_with_agent_install_status: ImageWithAgentInstallStatus

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.image_with_agent_install_status.image.id)



from .assign_role import AssignRoleAction, AssignRoleActionResult
from .check_permission import CheckPermissionAction, CheckPermissionActionResult
from .create_role import CreateRoleAction, CreateRoleActionResult
from .delete_role import DeleteRoleAction, DeleteRoleActionResult
from .update_role import UpdateRoleAction, UpdateRoleActionResult

__all__ = [
    "AssignRoleAction",
    "AssignRoleActionResult",
    "CheckPermissionAction",
    "CheckPermissionActionResult",
    "CreateRoleAction",
    "CreateRoleActionResult",
    "DeleteRoleAction",
    "DeleteRoleActionResult",
    "UpdateRoleAction",
    "UpdateRoleActionResult",
]



"""Repository pattern implementation for schedule operations."""

import logging
from typing import Mapping, Optional

from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.common.types import SessionId, SlotName, SlotTypes
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

from .cache_source.cache_source import ScheduleCacheSource
from .db_source.db_source import ScheduleDBSource
from .types.allocation import AllocationBatch
from .types.base import SchedulingSpec
from .types.scheduling import SchedulingData
from .types.session import (
    MarkTerminatingResult,
    SessionTerminationResult,
    SweptSessionInfo,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


schedule_repository_resilience = Resilience(
    policies=[
        MetricPolicy(MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.SCHEDULE_REPOSITORY)),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class ScheduleRepository:
    """
    Repository that orchestrates between DB and cache sources for scheduling operations.
    """

    _db_source: ScheduleDBSource
    _cache_source: ScheduleCacheSource
    _config_provider: ManagerConfigProvider

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        valkey_stat: ValkeyStatClient,
        config_provider: ManagerConfigProvider,
    ):
        self._db_source = ScheduleDBSource(db)
        self._cache_source = ScheduleCacheSource(valkey_stat)
        self._config_provider = config_provider

    @schedule_repository_resilience.apply()
    async def get_scheduling_data(self, scaling_group: str) -> Optional[SchedulingData]:
        """
        Get scheduling data from database.
        Returns None if no pending sessions exist.
        Raises ScalingGroupNotFound if scaling group doesn't exist.
        """
        # Get scheduling specification
        known_slot_types = await self._get_known_slot_types()
        max_container_count = await self._get_max_container_count()
        spec = SchedulingSpec(
            known_slot_types=known_slot_types,
            max_container_count=max_container_count,
        )

        # Fetch data from DB (will raise ScalingGroupNotFound if not found)
        scheduling_data = await self._db_source.get_scheduling_data(scaling_group, spec)
        if not scheduling_data.pending_sessions.sessions:
            return None

        return scheduling_data

    @schedule_repository_resilience.apply()
    async def allocate_sessions(self, allocation_batch: AllocationBatch) -> None:
        """
        Allocate sessions by updating DB.
        Agent occupied slots are synced directly in the DB.
        """
        # Update DB
        await self._db_source.allocate_sessions(allocation_batch)

    @schedule_repository_resilience.apply()
    async def get_pending_timeout_sessions(self) -> list[SweptSessionInfo]:
        """
        Get sessions that have exceeded their pending timeout.
        The timeout is determined by each scaling group's scheduler_opts.
        """
        # Fetch from DB source
        return await self._db_source.get_pending_timeout_sessions()

    @schedule_repository_resilience.apply()
    async def batch_update_terminated_status(
        self,
        session_results: list[SessionTerminationResult],
    ) -> None:
        """
        Update terminated status in DB.
        Agent occupied slots are synced directly in the DB.
        """
        if not session_results:
            return

        # Update DB
        await self._db_source.batch_update_terminated_status(session_results)

    @schedule_repository_resilience.apply()
    async def mark_sessions_terminating(
        self, session_ids: list[SessionId], reason: str = "USER_REQUESTED"
    ) -> MarkTerminatingResult:
        """
        Mark sessions for termination.
        """
        # Delegate to DB source
        return await self._db_source.mark_sessions_terminating(session_ids, reason)

    async def _get_known_slot_types(self) -> Mapping[SlotName, SlotTypes]:
        """
        Get known slot types from configuration.
        """
        return await self._config_provider.legacy_etcd_config_loader.get_resource_slots()

    async def _get_max_container_count(self) -> Optional[int]:
        """
        Get max container count from configuration.
        """
        raw_value = await self._config_provider.legacy_etcd_config_loader.get_raw(
            "config/agent/max-container-count"
        )
        return int(raw_value) if raw_value else None



import uuid
from dataclasses import dataclass
from typing import Any, Optional, override

from ai.backend.common.types import AccessKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.session.base import SessionAction


@dataclass
class GetStatusHistoryAction(SessionAction):
    session_name: str
    owner_access_key: AccessKey

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "get_status_history"


@dataclass
class GetStatusHistoryActionResult(BaseActionResult):
    status_history: dict[str, Any]
    session_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.session_id)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.notification import NotificationRuleData, NotificationRuleModifier

from .base import NotificationAction


@dataclass
class UpdateRuleAction(NotificationAction):
    """Action to update a notification rule."""

    rule_id: UUID
    modifier: NotificationRuleModifier

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "update_rule"

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.rule_id)


@dataclass
class UpdateRuleActionResult(BaseActionResult):
    """Result of updating a notification rule."""

    rule_data: NotificationRuleData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.rule_data.id)



"""
REST API handlers for notification system.
Provides CRUD endpoints for notification channels and rules.
"""

from __future__ import annotations

from http import HTTPStatus
from typing import Iterable, Tuple

import aiohttp_cors
from aiohttp import web

from ai.backend.common.api_handlers import APIResponse, BodyParam, PathParam, api_handler
from ai.backend.common.contexts.user import current_user
from ai.backend.common.dto.manager.notification import (
    CreateNotificationChannelRequest,
    CreateNotificationChannelResponse,
    CreateNotificationRuleRequest,
    CreateNotificationRuleResponse,
    DeleteNotificationChannelResponse,
    DeleteNotificationRuleResponse,
    GetNotificationChannelResponse,
    GetNotificationRuleResponse,
    ListNotificationChannelsResponse,
    ListNotificationRulesResponse,
    ListNotificationRuleTypesResponse,
    NotificationRuleType,
    NotificationRuleTypeSchemaResponse,
    PaginationInfo,
    SearchNotificationChannelsRequest,
    SearchNotificationRulesRequest,
    UpdateNotificationChannelRequest,
    UpdateNotificationChannelResponse,
    UpdateNotificationRuleRequest,
    UpdateNotificationRuleResponse,
    ValidateNotificationChannelRequest,
    ValidateNotificationChannelResponse,
    ValidateNotificationRuleRequest,
    ValidateNotificationRuleResponse,
)
from ai.backend.manager.dto.context import ProcessorsCtx
from ai.backend.manager.dto.notification_request import (
    DeleteNotificationChannelPathParam,
    DeleteNotificationRulePathParam,
    GetNotificationChannelPathParam,
    GetNotificationRulePathParam,
    RuleTypePathParam,
    UpdateNotificationChannelPathParam,
    UpdateNotificationRulePathParam,
    ValidateNotificationChannelPathParam,
    ValidateNotificationRulePathParam,
)
from ai.backend.manager.services.notification.actions import (
    CreateChannelAction,
    CreateRuleAction,
    DeleteChannelAction,
    DeleteRuleAction,
    GetChannelAction,
    GetRuleAction,
    SearchChannelsAction,
    SearchRulesAction,
    UpdateChannelAction,
    UpdateRuleAction,
    ValidateChannelAction,
    ValidateRuleAction,
)

from ...data.notification import NotificationChannelCreator, NotificationRuleCreator
from ..auth import auth_required_for_method
from ..types import CORSOptions, WebMiddleware
from .adapter import NotificationChannelAdapter, NotificationRuleAdapter

__all__ = ("create_app",)


class NotificationAPIHandler:
    """REST API handler class for notification operations."""

    def __init__(self) -> None:
        self.channel_adapter = NotificationChannelAdapter()
        self.rule_adapter = NotificationRuleAdapter()

    # Notification Channel Endpoints

    @auth_required_for_method
    @api_handler
    async def create_channel(
        self,
        body: BodyParam[CreateNotificationChannelRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Create a new notification channel."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can create notification channels.")

        # Convert request to creator
        # config validator in request DTO ensures this is WebhookConfig

        creator = NotificationChannelCreator(
            name=body.parsed.name,
            description=body.parsed.description,
            channel_type=body.parsed.channel_type,
            config=body.parsed.config,
            enabled=body.parsed.enabled,
            created_by=me.user_id,
        )

        # Call service action
        action_result = await processors.notification.create_channel.wait_for_complete(
            CreateChannelAction(creator=creator)
        )

        # Build response
        resp = CreateNotificationChannelResponse(
            channel=self.channel_adapter.convert_to_dto(action_result.channel_data)
        )
        return APIResponse.build(status_code=HTTPStatus.CREATED, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def search_channels(
        self,
        body: BodyParam[SearchNotificationChannelsRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search notification channels with filters, orders, and pagination."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can search notification channels.")

        # Build querier using adapter
        querier = self.channel_adapter.build_querier(body.parsed)

        # Call service action
        action_result = await processors.notification.search_channels.wait_for_complete(
            SearchChannelsAction(querier=querier)
        )

        # Build response
        resp = ListNotificationChannelsResponse(
            channels=[self.channel_adapter.convert_to_dto(ch) for ch in action_result.channels],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset or 0,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def get_channel(
        self,
        path: PathParam[GetNotificationChannelPathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Get a specific notification channel."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can get notification channels.")

        # Call service action
        action_result = await processors.notification.get_channel.wait_for_complete(
            GetChannelAction(channel_id=path.parsed.channel_id)
        )

        # Build response
        resp = GetNotificationChannelResponse(
            channel=self.channel_adapter.convert_to_dto(action_result.channel_data)
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def update_channel(
        self,
        path: PathParam[UpdateNotificationChannelPathParam],
        body: BodyParam[UpdateNotificationChannelRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Update an existing notification channel."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can update notification channels.")

        # Call service action
        action_result = await processors.notification.update_channel.wait_for_complete(
            UpdateChannelAction(
                channel_id=path.parsed.channel_id,
                modifier=self.channel_adapter.build_modifier(body.parsed),
            )
        )

        # Build response
        resp = UpdateNotificationChannelResponse(
            channel=self.channel_adapter.convert_to_dto(action_result.channel_data)
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def delete_channel(
        self,
        path: PathParam[DeleteNotificationChannelPathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Delete a notification channel."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can delete notification channels.")

        # Call service action
        action_result = await processors.notification.delete_channel.wait_for_complete(
            DeleteChannelAction(channel_id=path.parsed.channel_id)
        )

        # Build response
        resp = DeleteNotificationChannelResponse(deleted=action_result.deleted)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def validate_channel(
        self,
        path: PathParam[ValidateNotificationChannelPathParam],
        body: BodyParam[ValidateNotificationChannelRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Validate a notification channel by sending a test webhook."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can validate notification channels.")

        # Call service action
        await processors.notification.validate_channel.wait_for_complete(
            ValidateChannelAction(
                channel_id=path.parsed.channel_id,
                test_message=body.parsed.test_message,
            )
        )

        # Build response
        resp = ValidateNotificationChannelResponse(
            channel_id=path.parsed.channel_id,
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def validate_rule(
        self,
        path: PathParam[ValidateNotificationRulePathParam],
        body: BodyParam[ValidateNotificationRuleRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Validate a notification rule by rendering its template with test data."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can validate notification rules.")

        # Service layer handles rule fetching and data validation
        action_result = await processors.notification.validate_rule.wait_for_complete(
            ValidateRuleAction(
                rule_id=path.parsed.rule_id,
                notification_data=body.parsed.notification_data,
            )
        )

        # Build response
        resp = ValidateNotificationRuleResponse(
            message=action_result.message,
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def list_rule_types(
        self,
    ) -> APIResponse:
        """List all available notification rule types."""
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can list notification rule types.")

        # Return all available rule types from the enum
        resp = ListNotificationRuleTypesResponse(rule_types=list(NotificationRuleType))
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def get_rule_type_schema(
        self,
        path: PathParam[RuleTypePathParam],
    ) -> APIResponse:
        """Get JSON schema for a notification rule type's message format."""
        from ai.backend.common.data.notification import NotifiableMessage

        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(
                reason="Only superadmin can view notification rule type schemas."
            )

        # Get schema for the requested rule type
        schema = NotifiableMessage.get_message_schema(path.parsed.rule_type)

        resp = NotificationRuleTypeSchemaResponse(
            rule_type=path.parsed.rule_type,
            json_schema=schema,
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Notification Rule Endpoints

    @auth_required_for_method
    @api_handler
    async def create_rule(
        self,
        body: BodyParam[CreateNotificationRuleRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Create a new notification rule."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can create notification rules.")

        # Convert request to creator
        creator = NotificationRuleCreator(
            name=body.parsed.name,
            description=body.parsed.description,
            rule_type=body.parsed.rule_type,
            channel_id=body.parsed.channel_id,
            message_template=body.parsed.message_template,
            enabled=body.parsed.enabled,
            created_by=me.user_id,
        )

        # Call service action
        action_result = await processors.notification.create_rule.wait_for_complete(
            CreateRuleAction(creator=creator)
        )

        # Build response
        resp = CreateNotificationRuleResponse(
            rule=self.rule_adapter.convert_to_dto(action_result.rule_data)
        )
        return APIResponse.build(status_code=HTTPStatus.CREATED, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def search_rules(
        self,
        body: BodyParam[SearchNotificationRulesRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search notification rules with filters, orders, and pagination."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can search notification rules.")

        # Build querier using adapter
        querier = self.rule_adapter.build_querier(body.parsed)

        # Call service action
        action_result = await processors.notification.search_rules.wait_for_complete(
            SearchRulesAction(querier=querier)
        )

        # Build response
        resp = ListNotificationRulesResponse(
            rules=[self.rule_adapter.convert_to_dto(rule) for rule in action_result.rules],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset or 0,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def get_rule(
        self,
        path: PathParam[GetNotificationRulePathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Get a specific notification rule."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can get notification rules.")

        # Call service action
        action_result = await processors.notification.get_rule.wait_for_complete(
            GetRuleAction(rule_id=path.parsed.rule_id)
        )

        # Build response
        resp = GetNotificationRuleResponse(
            rule=self.rule_adapter.convert_to_dto(action_result.rule_data)
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def update_rule(
        self,
        path: PathParam[UpdateNotificationRulePathParam],
        body: BodyParam[UpdateNotificationRuleRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Update an existing notification rule."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can update notification rules.")

        # Call service action
        action_result = await processors.notification.update_rule.wait_for_complete(
            UpdateRuleAction(
                rule_id=path.parsed.rule_id,
                modifier=self.rule_adapter.build_modifier(body.parsed),
            )
        )

        # Build response
        resp = UpdateNotificationRuleResponse(
            rule=self.rule_adapter.convert_to_dto(action_result.rule_data)
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def delete_rule(
        self,
        path: PathParam[DeleteNotificationRulePathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Delete a notification rule."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can delete notification rules.")

        # Call service action
        action_result = await processors.notification.delete_rule.wait_for_complete(
            DeleteRuleAction(rule_id=path.parsed.rule_id)
        )

        # Build response
        resp = DeleteNotificationRuleResponse(deleted=action_result.deleted)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    """Create aiohttp application for notification API endpoints."""
    app = web.Application()
    app["api_versions"] = (4, 5)
    app["prefix"] = "notifications"

    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    api_handler = NotificationAPIHandler()

    # Channel routes
    cors.add(app.router.add_route("POST", "/channels", api_handler.create_channel))
    cors.add(app.router.add_route("POST", "/channels/search", api_handler.search_channels))
    cors.add(app.router.add_route("GET", "/channels/{channel_id}", api_handler.get_channel))
    cors.add(app.router.add_route("PATCH", "/channels/{channel_id}", api_handler.update_channel))
    cors.add(app.router.add_route("DELETE", "/channels/{channel_id}", api_handler.delete_channel))
    cors.add(
        app.router.add_route(
            "POST", "/channels/{channel_id}/validate", api_handler.validate_channel
        )
    )

    # Rule routes
    cors.add(app.router.add_route("GET", "/rule-types", api_handler.list_rule_types))
    cors.add(
        app.router.add_route(
            "GET", "/rule-types/{rule_type}/schema", api_handler.get_rule_type_schema
        )
    )
    cors.add(app.router.add_route("POST", "/rules", api_handler.create_rule))
    cors.add(app.router.add_route("POST", "/rules/search", api_handler.search_rules))
    cors.add(app.router.add_route("GET", "/rules/{rule_id}", api_handler.get_rule))
    cors.add(app.router.add_route("PATCH", "/rules/{rule_id}", api_handler.update_rule))
    cors.add(app.router.add_route("DELETE", "/rules/{rule_id}", api_handler.delete_rule))
    cors.add(app.router.add_route("POST", "/rules/{rule_id}/validate", api_handler.validate_rule))

    return app, []



import uuid
from dataclasses import dataclass
from typing import Any, Optional

from ai.backend.common.container_registry import ContainerRegistryType


@dataclass
class ContainerRegistryData:
    id: uuid.UUID
    url: str
    registry_name: str
    type: ContainerRegistryType
    project: Optional[str]
    username: Optional[str]
    password: Optional[str]
    ssl_verify: Optional[bool]
    is_global: Optional[bool]
    # TODO: Add proper type
    extra: Optional[dict[str, Any]]



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.artifact_registry.actions.base import ArtifactRegistryAction


@dataclass
class DeleteHuggingFaceRegistryAction(ArtifactRegistryAction):
    registry_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.registry_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "delete_huggingface_registry"


@dataclass
class DeleteHuggingFaceRegistryActionResult(BaseActionResult):
    deleted_registry_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.deleted_registry_id)



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from ai.backend.common.dependencies import NonMonitorableDependencyProvider
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.logging.types import LogLevel
from ai.backend.manager.config.loader.config_overrider import ConfigOverrider
from ai.backend.manager.config.loader.etcd_loader import (
    EtcdCommonConfigLoader,
    EtcdManagerConfigLoader,
)
from ai.backend.manager.config.loader.legacy_etcd_loader import LegacyEtcdLoader
from ai.backend.manager.config.loader.loader_chain import LoaderChain
from ai.backend.manager.config.loader.toml_loader import TomlConfigLoader
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.config.watchers.etcd import EtcdConfigWatcher


@dataclass
class ConfigProviderInput:
    """Input required for config provider setup.

    Contains etcd client and configuration parameters.
    """

    etcd: AsyncEtcd
    config_path: Path
    log_level: LogLevel = LogLevel.NOTSET


class ConfigProviderDependency(
    NonMonitorableDependencyProvider[ConfigProviderInput, ManagerConfigProvider]
):
    """Provides ManagerConfigProvider lifecycle management.

    Creates the config provider with file, env, etcd-based loaders and watchers.
    """

    @property
    def stage_name(self) -> str:
        return "config-provider"

    @asynccontextmanager
    async def provide(
        self, setup_input: ConfigProviderInput
    ) -> AsyncIterator[ManagerConfigProvider]:
        """Initialize and provide a config provider.

        Args:
            setup_input: Input containing etcd client and config parameters

        Yields:
            Initialized ManagerConfigProvider
        """
        # Create loader chain following server.py pattern
        loaders: list = []

        # Add file loader if config path is provided
        if setup_input.config_path:
            toml_loader = TomlConfigLoader(setup_input.config_path, "manager")
            loaders.append(toml_loader)

        # Add legacy etcd loader
        legacy_etcd_loader = LegacyEtcdLoader(setup_input.etcd)
        loaders.append(legacy_etcd_loader)

        # Add etcd config loaders
        loaders.append(EtcdCommonConfigLoader(setup_input.etcd))
        loaders.append(EtcdManagerConfigLoader(setup_input.etcd))

        # Add overrides
        overrides: list[tuple[tuple[str, ...], Any]] = [
            (("debug", "enabled"), setup_input.log_level == LogLevel.DEBUG),
        ]
        if setup_input.log_level != LogLevel.NOTSET:
            overrides += [
                (("logging", "level"), setup_input.log_level),
                (("logging", "pkg-ns", "ai.backend"), setup_input.log_level),
            ]
        loaders.append(ConfigOverrider(overrides))

        # Create unified loader chain
        unified_loader = LoaderChain(loaders)
        etcd_watcher = EtcdConfigWatcher(setup_input.etcd)

        # Create the config provider
        config_provider = await ManagerConfigProvider.create(
            unified_loader,
            etcd_watcher,
            legacy_etcd_loader,
        )

        try:
            yield config_provider
        finally:
            await config_provider.terminate()



import asyncio
import base64
import functools
import logging
import secrets
import uuid
from dataclasses import asdict, dataclass
from datetime import datetime
from typing import Any, Mapping, MutableMapping, Optional, Union, cast
from urllib.parse import urlparse

import aiohttp
import aiotools
import multidict
import trafaret as t
from aiohttp.multipart import BodyPartReader
from dateutil.tz import tzutc

from ai.backend.common.bgtask.bgtask import BackgroundTaskManager
from ai.backend.common.events.event_types.kernel.types import KernelLifecycleEventReason
from ai.backend.common.events.fetcher import EventFetcher
from ai.backend.common.events.hub.hub import EventHub
from ai.backend.common.exception import (
    BackendAIError,
    InvalidAPIParameters,
    UnknownImageReference,
)
from ai.backend.common.json import load_json
from ai.backend.common.plugin.monitor import ErrorPluginContext
from ai.backend.common.types import (
    ImageAlias,
    SessionId,
    SessionTypes,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.api.scaling_group import query_wsproxy_status
from ai.backend.manager.api.session import (
    CustomizedImageVisibilityScope,
    drop_undefined,
    overwritten_param_check,
)
from ai.backend.manager.api.utils import undefined
from ai.backend.manager.bgtask.tasks.commit_session import CommitSessionManifest
from ai.backend.manager.bgtask.types import ManagerBgtaskName
from ai.backend.manager.data.image.types import ImageIdentifier
from ai.backend.manager.data.session.types import SessionStatus
from ai.backend.manager.errors.common import (
    InternalServerError,
    ServiceUnavailable,
)
from ai.backend.manager.errors.image import UnknownImageReferenceError
from ai.backend.manager.errors.kernel import (
    InvalidSessionData,
    KernelNotReady,
    QuotaExceeded,
    SessionAlreadyExists,
    SessionNotFound,
    TooManySessionsMatched,
)
from ai.backend.manager.errors.resource import (
    AppNotFound,
    NoCurrentTaskContext,
    TaskTemplateNotFound,
)
from ai.backend.manager.errors.storage import VFolderBadRequest
from ai.backend.manager.idle import IdleCheckerHost
from ai.backend.manager.models.group import GroupRow
from ai.backend.manager.models.session import (
    DEAD_SESSION_STATUSES,
    PRIVATE_SESSION_TYPES,
    KernelLoadingStrategy,
)
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.repositories.session.admin_repository import AdminSessionRepository
from ai.backend.manager.repositories.session.repository import SessionRepository
from ai.backend.manager.services.session.actions.check_and_transit_status import (
    CheckAndTransitStatusAction,
    CheckAndTransitStatusActionResult,
    CheckAndTransitStatusBatchAction,
    CheckAndTransitStatusBatchActionResult,
)
from ai.backend.manager.services.session.actions.commit_session import (
    CommitSessionAction,
    CommitSessionActionResult,
)
from ai.backend.manager.services.session.actions.complete import (
    CompleteAction,
    CompleteActionResult,
)
from ai.backend.manager.services.session.actions.convert_session_to_image import (
    ConvertSessionToImageAction,
    ConvertSessionToImageActionResult,
)
from ai.backend.manager.services.session.actions.create_cluster import (
    CreateClusterAction,
    CreateClusterActionResult,
)
from ai.backend.manager.services.session.actions.create_from_params import (
    CreateFromParamsAction,
    CreateFromParamsActionResult,
)
from ai.backend.manager.services.session.actions.create_from_template import (
    CreateFromTemplateAction,
    CreateFromTemplateActionResult,
)
from ai.backend.manager.services.session.actions.destroy_session import (
    DestroySessionAction,
    DestroySessionActionResult,
)
from ai.backend.manager.services.session.actions.download_file import (
    DownloadFileAction,
    DownloadFileActionResult,
)
from ai.backend.manager.services.session.actions.download_files import (
    DownloadFilesAction,
    DownloadFilesActionResult,
)
from ai.backend.manager.services.session.actions.execute_session import (
    ExecuteSessionAction,
    ExecuteSessionActionResult,
)
from ai.backend.manager.services.session.actions.get_abusing_report import (
    GetAbusingReportAction,
    GetAbusingReportActionResult,
)
from ai.backend.manager.services.session.actions.get_commit_status import (
    GetCommitStatusAction,
    GetCommitStatusActionResult,
)
from ai.backend.manager.services.session.actions.get_container_logs import (
    GetContainerLogsAction,
    GetContainerLogsActionResult,
)
from ai.backend.manager.services.session.actions.get_dependency_graph import (
    GetDependencyGraphAction,
    GetDependencyGraphActionResult,
)
from ai.backend.manager.services.session.actions.get_direct_access_info import (
    GetDirectAccessInfoAction,
    GetDirectAccessInfoActionResult,
)
from ai.backend.manager.services.session.actions.get_session_info import (
    GetSessionInfoAction,
    GetSessionInfoActionResult,
)
from ai.backend.manager.services.session.actions.get_status_history import (
    GetStatusHistoryAction,
    GetStatusHistoryActionResult,
)
from ai.backend.manager.services.session.actions.interrupt_session import (
    InterruptSessionAction,
    InterruptSessionActionResult,
)
from ai.backend.manager.services.session.actions.list_files import (
    ListFilesAction,
    ListFilesActionResult,
)
from ai.backend.manager.services.session.actions.match_sessions import (
    MatchSessionsAction,
    MatchSessionsActionResult,
)
from ai.backend.manager.services.session.actions.modify_session import (
    ModifySessionAction,
    ModifySessionActionResult,
)
from ai.backend.manager.services.session.actions.rename_session import (
    RenameSessionAction,
    RenameSessionActionResult,
)
from ai.backend.manager.services.session.actions.restart_session import (
    RestartSessionAction,
    RestartSessionActionResult,
)
from ai.backend.manager.services.session.actions.shutdown_service import (
    ShutdownServiceAction,
    ShutdownServiceActionResult,
)
from ai.backend.manager.services.session.actions.start_service import (
    StartServiceAction,
    StartServiceActionResult,
)
from ai.backend.manager.services.session.actions.upload_files import (
    UploadFilesAction,
    UploadFilesActionResult,
)
from ai.backend.manager.services.session.types import CommitStatusInfo, LegacySessionInfo
from ai.backend.manager.sokovan.scheduling_controller import SchedulingController
from ai.backend.manager.types import UserScope

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore


@dataclass
class SessionServiceArgs:
    agent_registry: AgentRegistry
    event_fetcher: EventFetcher
    background_task_manager: BackgroundTaskManager
    event_hub: EventHub
    error_monitor: ErrorPluginContext
    idle_checker_host: IdleCheckerHost
    session_repository: SessionRepository
    admin_session_repository: AdminSessionRepository
    scheduling_controller: SchedulingController


class SessionService:
    _agent_registry: AgentRegistry
    _background_task_manager: BackgroundTaskManager
    _event_fetcher: EventFetcher
    _event_hub: EventHub
    _error_monitor: ErrorPluginContext
    _idle_checker_host: IdleCheckerHost
    _session_repository: SessionRepository
    _admin_session_repository: AdminSessionRepository
    _scheduling_controller: SchedulingController
    _database_ptask_group: aiotools.PersistentTaskGroup
    _rpc_ptask_group: aiotools.PersistentTaskGroup

    def __init__(
        self,
        args: SessionServiceArgs,
    ) -> None:
        self._agent_registry = args.agent_registry
        self._event_hub = args.event_hub
        self._event_fetcher = args.event_fetcher
        self._background_task_manager = args.background_task_manager
        self._error_monitor = args.error_monitor
        self._idle_checker_host = args.idle_checker_host
        self._session_repository = args.session_repository
        self._admin_session_repository = args.admin_session_repository
        self._scheduling_controller = args.scheduling_controller
        self._database_ptask_group = aiotools.PersistentTaskGroup()
        self._rpc_ptask_group = aiotools.PersistentTaskGroup()
        self._webhook_ptask_group = aiotools.PersistentTaskGroup()

    async def commit_session(self, action: CommitSessionAction) -> CommitSessionActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        filename = action.filename

        myself = asyncio.current_task()
        if myself is None:
            raise NoCurrentTaskContext("No current asyncio task context available")

        session = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )

        resp: Mapping[str, Any] = await asyncio.shield(
            self._rpc_ptask_group.create_task(
                self._agent_registry.commit_session_to_file(session, filename),
            ),
        )

        return CommitSessionActionResult(
            session_data=session.to_dataclass(),
            commit_result=resp,
        )

    async def complete(self, action: CompleteAction) -> CompleteActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        code = action.code
        options = action.options or {}

        session = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )
        try:
            await self._agent_registry.increment_session_usage(session)
            resp = await self._agent_registry.get_completions(session, code, opts=options)
        except AssertionError:
            raise InvalidAPIParameters
        return CompleteActionResult(
            session_data=session.to_dataclass(),
            result=resp,
        )

    async def convert_session_to_image(
        self, action: ConvertSessionToImageAction
    ) -> ConvertSessionToImageActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        image_name = action.image_name
        image_visibility = action.image_visibility
        image_owner_id = action.image_owner_id

        myself = asyncio.current_task()
        if myself is None:
            raise NoCurrentTaskContext("No current asyncio task context available")

        if image_visibility != CustomizedImageVisibilityScope.USER:
            raise InvalidAPIParameters(f"Unsupported visibility scope {image_visibility}")

        # check if user has passed its limit of customized image count
        existing_image_count = await self._session_repository.get_customized_image_count(
            image_visibility.value, str(image_owner_id)
        )
        customized_image_count_limit = action.max_customized_image_count
        if customized_image_count_limit <= existing_image_count:
            raise QuotaExceeded(
                extra_msg=(
                    "You have reached your customized image count quota. "
                    f"(current: {existing_image_count}, limit: {customized_image_count_limit})"
                ),
                extra_data={
                    "limit": customized_image_count_limit,
                    "current": existing_image_count,
                },
            )

        session = await self._session_repository.get_session_with_group(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )

        project: GroupRow = session.group
        if not project.container_registry:
            raise InvalidAPIParameters(
                "Project not ready to convert session image (registry configuration not populated)"
            )

        registry_hostname = project.container_registry["registry"]
        registry_project = project.container_registry["project"]

        registry_conf = await self._session_repository.get_container_registry(
            registry_hostname, registry_project
        )
        if not registry_conf:
            raise InvalidAPIParameters(
                f"Project {registry_project} not found in registry {registry_hostname}."
            )

        # Validate image exists
        await self._session_repository.resolve_image([
            ImageIdentifier(session.main_kernel.image, session.main_kernel.architecture)
        ])

        # Create manifest for background task
        manifest = CommitSessionManifest(
            session_id=session.id,
            registry_hostname=registry_hostname,
            registry_project=registry_project,
            image_name=image_name,
            image_visibility=image_visibility,
            image_owner_id=str(image_owner_id),
            user_email=action.user_email,
        )

        task_id = await self._background_task_manager.start_retriable(
            ManagerBgtaskName.COMMIT_SESSION,
            manifest,
        )

        return ConvertSessionToImageActionResult(
            task_id=task_id, session_data=session.to_dataclass()
        )

    async def create_cluster(self, action: CreateClusterAction) -> CreateClusterActionResult:
        template_id = action.template_id
        user_id = action.user_id
        user_role = action.user_role
        sudo_session_enabled = action.sudo_session_enabled
        keypair_resource_policy = action.keypair_resource_policy
        requester_access_key = action.requester_access_key
        owner_access_key = action.owner_access_key
        domain_name = action.domain_name
        group_name = action.group_name
        scaling_group_name = action.scaling_group_name
        session_name = action.session_name
        session_type = action.session_type
        enqueue_only = action.enqueue_only
        max_wait_seconds = action.max_wait_seconds
        tag = action.tag

        template = await self._session_repository.get_template_by_id(template_id)
        log.debug("task template: {}", template)
        if not template:
            raise TaskTemplateNotFound

        try:
            _, group_id, resource_policy = await self._session_repository.query_userinfo(
                user_id,
                requester_access_key,
                user_role,
                domain_name,
                keypair_resource_policy,
                domain_name,
                group_name,
                query_on_behalf_of=(None if owner_access_key is undefined else owner_access_key),
            )
        except ValueError as e:
            raise InvalidAPIParameters(str(e))

        try:
            resp = await self._agent_registry.create_cluster(
                template,
                session_name,
                UserScope(
                    domain_name=domain_name,
                    group_id=group_id,
                    user_uuid=user_id,
                    user_role=user_role,
                ),
                owner_access_key,
                resource_policy,
                scaling_group_name,
                session_type,
                tag,
                enqueue_only=enqueue_only,
                max_wait_seconds=max_wait_seconds,
                sudo_session_enabled=sudo_session_enabled,
            )
            return CreateClusterActionResult(result=resp, session_id=resp["kernelId"])
        except TooManySessionsMatched:
            raise SessionAlreadyExists
        except BackendAIError:
            raise
        except UnknownImageReference:
            raise UnknownImageReferenceError("Unknown image reference!")
        except Exception as e:
            await self._error_monitor.capture_exception()
            log.exception("GET_OR_CREATE: unexpected error!", e)
            raise InternalServerError

    async def create_from_params(
        self, action: CreateFromParamsAction
    ) -> CreateFromParamsActionResult:
        user_id = action.user_id
        user_role = action.user_role
        sudo_session_enabled = action.sudo_session_enabled
        keypair_resource_policy = action.keypair_resource_policy
        requester_access_key = action.requester_access_key

        owner_access_key = action.params.owner_access_key
        domain_name = action.params.domain_name
        group_name = action.params.group_name
        config = action.params.config
        cluster_size = action.params.cluster_size
        cluster_mode = action.params.cluster_mode
        session_name = action.params.session_name
        session_type = action.params.session_type
        enqueue_only = action.params.enqueue_only
        max_wait_seconds = action.params.max_wait_seconds
        tag = action.params.tag
        image = action.params.image
        architecture = action.params.architecture
        priority = action.params.priority
        bootstrap_script = action.params.bootstrap_script
        dependencies = action.params.dependencies
        startup_command = action.params.startup_command
        starts_at = action.params.starts_at
        batch_timeout = action.params.batch_timeout
        callback_url = action.params.callback_url
        reuse_if_exists = action.params.reuse_if_exists

        owner_uuid, group_id, resource_policy = await self._session_repository.query_userinfo(
            user_id,
            requester_access_key,
            user_role,
            domain_name,
            keypair_resource_policy,
            domain_name,
            group_name,
            query_on_behalf_of=(None if owner_access_key is undefined else owner_access_key),
        )

        try:
            image_row = await self._session_repository.resolve_image([
                ImageIdentifier(
                    image,
                    architecture,
                ),
                ImageAlias(image),
            ])

            resp = await self._agent_registry.create_session(
                session_name,
                image_row.image_ref,
                UserScope(
                    domain_name=domain_name,
                    group_id=group_id,
                    user_uuid=user_id,
                    user_role=user_role,
                ),
                owner_access_key,
                resource_policy,
                session_type,
                config,
                cluster_mode,
                cluster_size,
                reuse=reuse_if_exists,
                priority=priority,
                enqueue_only=enqueue_only,
                max_wait_seconds=max_wait_seconds,
                bootstrap_script=bootstrap_script,
                dependencies=dependencies,
                startup_command=startup_command,
                starts_at_timestamp=starts_at,
                batch_timeout=batch_timeout,
                tag=tag,
                callback_url=callback_url,
                sudo_session_enabled=sudo_session_enabled,
            )
            return CreateFromParamsActionResult(
                session_id=uuid.UUID(resp["sessionId"]), result=resp
            )
        except UnknownImageReference:
            raise UnknownImageReferenceError(f"Unknown image reference: {image}")
        except BackendAIError:
            raise
        except Exception as e:
            await self._error_monitor.capture_exception(context={"user": owner_uuid})
            log.exception("GET_OR_CREATE: unexpected error!", e)
            raise InternalServerError

    async def create_from_template(
        self, action: CreateFromTemplateAction
    ) -> CreateFromTemplateActionResult:
        template_id = action.params.template_id
        user_id = action.user_id
        user_role = action.user_role
        sudo_session_enabled = action.sudo_session_enabled
        keypair_resource_policy = action.keypair_resource_policy
        requester_access_key = action.requester_access_key

        template_info = await self._session_repository.get_template_info_by_id(template_id)
        if not template_info:
            raise TaskTemplateNotFound
        template = template_info["template"]

        group_name = None
        if template_info["domain_name"] and template_info["group_id"]:
            group_name = await self._session_repository.get_group_name_by_domain_and_id(
                template_info["domain_name"], template_info["group_id"]
            )

        if isinstance(template, str):
            template = load_json(template)

        log.debug("Template: {0}", template)

        param_from_template = {
            "image": template["spec"]["kernel"]["image"],
            "architecture": template["spec"]["kernel"]["architecture"],
        }

        if "domain_name" in template_info:
            param_from_template["domain"] = template_info["domain_name"]
        if group_name:
            param_from_template["group"] = group_name
        if template["spec"]["session_type"] == "interactive":
            param_from_template["session_type"] = SessionTypes.INTERACTIVE
        elif template["spec"]["session_type"] == "batch":
            param_from_template["session_type"] = SessionTypes.BATCH
        elif template["spec"]["session_type"] == "inference":
            param_from_template["session_type"] = SessionTypes.INFERENCE

        if tag := template["metadata"].get("tag"):
            param_from_template["tag"] = tag
        if runtime_opt := template["spec"]["kernel"]["run"]:
            if bootstrap := runtime_opt["bootstrap"]:
                param_from_template["bootstrap_script"] = bootstrap
            if startup := runtime_opt["startup_command"]:
                param_from_template["startup_command"] = startup

        config_from_template: MutableMapping[Any, Any] = {}
        if scaling_group := template["spec"].get("scaling_group"):
            config_from_template["scaling_group"] = scaling_group
        if mounts := template["spec"].get("mounts"):
            config_from_template["mounts"] = list(mounts.keys())
            config_from_template["mount_map"] = {
                key: value for (key, value) in mounts.items() if len(value) > 0
            }
        if environ := template["spec"]["kernel"].get("environ"):
            config_from_template["environ"] = environ
        if resources := template["spec"].get("resources"):
            config_from_template["resources"] = resources
        if "agent_list" in template["spec"]:
            config_from_template["agent_list"] = template["spec"]["agent_list"]

        override_config = drop_undefined(dict(action.params.config))
        override_params = drop_undefined(dict(asdict(action.params)))

        log.debug("Default config: {0}", config_from_template)
        log.debug("Default params: {0}", param_from_template)

        log.debug("Override config: {0}", override_config)
        log.debug("Override params: {0}", override_params)

        if override_config:
            config_from_template.update(override_config)
        if override_params:
            param_from_template.update(override_params)

        try:
            params = overwritten_param_check.check(param_from_template)
        except RuntimeError as e1:
            log.exception(e1)
            raise InvalidAPIParameters("Error while validating template")
        except t.DataError as e2:
            log.debug("Error: {0}", str(e2))
            raise InvalidAPIParameters("Error while validating template")
        params["config"] = config_from_template

        log.debug("Updated param: {0}", params)

        if git := template["spec"]["kernel"]["git"]:
            if _dest := git.get("dest_dir"):
                target = _dest
            else:
                target = git["repository"].split("/")[-1]

            cmd_builder = "git clone "
            if credential := git.get("credential"):
                proto, url = git["repository"].split("://")
                cmd_builder += f"{proto}://{credential['username']}:{credential['password']}@{url}"
            else:
                cmd_builder += git["repository"]
            if branch := git.get("branch"):
                cmd_builder += f" -b {branch}"
            cmd_builder += f" {target}\n"

            if commit := git.get("commit"):
                cmd_builder = "CWD=$(pwd)\n" + cmd_builder
                cmd_builder += f"cd {target}\n"
                cmd_builder += f"git checkout {commit}\n"
                cmd_builder += "cd $CWD\n"

            bootstrap = base64.b64decode(params.get("bootstrap_script") or b"").decode()
            bootstrap += "\n"
            bootstrap += cmd_builder
            params["bootstrap_script"] = base64.b64encode(bootstrap.encode()).decode()

        owner_access_key = params["owner_access_key"]
        config = params["config"]
        cluster_size = params["cluster_size"]
        cluster_mode = params["cluster_mode"]
        session_name = params["session_name"]
        session_type = params["session_type"]
        enqueue_only = params["enqueue_only"]
        max_wait_seconds = params["max_wait_seconds"]
        tag = params["tag"]
        image = params["image"]
        architecture = params["architecture"]
        priority = params["priority"]
        bootstrap_script = params["bootstrap_script"]
        dependencies = params["dependencies"]
        startup_command = params["startup_command"]
        starts_at = params["starts_at"]
        batch_timeout = params["batch_timeout"]
        callback_url = params["callback_url"]
        reuse_if_exists = params["reuse_if_exists"]
        domain_name = params["domain_name"]

        owner_uuid, group_id, resource_policy = await self._session_repository.query_userinfo(
            user_id,
            requester_access_key,
            user_role,
            domain_name,
            keypair_resource_policy,
            domain_name,
            params["group_name"],
            query_on_behalf_of=(None if owner_access_key is undefined else owner_access_key),
        )

        try:
            image_row = await self._session_repository.resolve_image([
                ImageIdentifier(
                    image,
                    architecture,
                ),
                ImageAlias(image),
            ])

            resp = await self._agent_registry.create_session(
                session_name,
                image_row.image_ref,
                UserScope(
                    domain_name=domain_name,
                    group_id=group_id,
                    user_uuid=user_id,
                    user_role=user_role,
                ),
                owner_access_key,
                resource_policy,
                session_type,
                config,
                cluster_mode,
                cluster_size,
                reuse=reuse_if_exists,
                priority=priority,
                enqueue_only=enqueue_only,
                max_wait_seconds=max_wait_seconds,
                bootstrap_script=bootstrap_script,
                dependencies=dependencies,
                startup_command=startup_command,
                starts_at_timestamp=starts_at,
                batch_timeout=batch_timeout,
                tag=tag,
                callback_url=callback_url,
                sudo_session_enabled=sudo_session_enabled,
            )
            return CreateFromTemplateActionResult(session_id=resp["sessionId"], result=resp)
        except UnknownImageReference:
            raise UnknownImageReferenceError(f"Unknown image reference: {image}")
        except BackendAIError:
            raise
        except Exception as e:
            await self._error_monitor.capture_exception(context={"user": owner_uuid})
            log.exception("GET_OR_CREATE: unexpected error!", e)
            raise InternalServerError

    async def destroy_session(self, action: DestroySessionAction) -> DestroySessionActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        forced = action.forced
        recursive = action.recursive

        # Get session IDs to terminate (based on recursive flag)
        session_ids = await self._session_repository.get_target_session_ids(
            session_name,
            owner_access_key,
            recursive=recursive,
        )

        # Determine termination reason based on forced flag
        if forced:
            reason = KernelLifecycleEventReason.FORCE_TERMINATED
        else:
            reason = KernelLifecycleEventReason.USER_REQUESTED

        # Mark sessions for termination
        mark_result = await self._scheduling_controller.mark_sessions_for_termination(
            session_ids,
            reason=reason.value,
        )

        # Build stats for response - prioritize cancelled over terminating
        if mark_result.cancelled_sessions:
            last_stat = {"status": "cancelled"}
        elif mark_result.terminating_sessions:
            last_stat = {"status": "terminated"}
        else:
            last_stat = {}

        # Return response - same format for both recursive and non-recursive
        resp = {"stats": last_stat}
        return DestroySessionActionResult(result=resp)

    async def download_file(self, action: DownloadFileAction) -> DownloadFileActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        user_id = action.user_id
        file = action.file
        try:
            session = await self._session_repository.get_session_validated(
                session_name,
                owner_access_key,
                kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
            )
            await self._agent_registry.increment_session_usage(session)
            result = await self._agent_registry.download_single(session, owner_access_key, file)
        except (ValueError, FileNotFoundError):
            raise InvalidAPIParameters("The file is not found.")
        except asyncio.CancelledError:
            raise
        except BackendAIError:
            raise
        except Exception as e:
            await self._error_monitor.capture_exception(context={"user": user_id})
            log.exception("DOWNLOAD_SINGLE: unexpected error!", e)
            raise InternalServerError

        return DownloadFileActionResult(bytes=result, session_data=session.to_dataclass())

    async def download_files(self, action: DownloadFilesAction) -> DownloadFilesActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        user_id = action.user_id
        files = action.files
        session = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )
        try:
            if len(files) > 5:
                raise VFolderBadRequest("Too many files (maximum 5 files allowed)")
            await self._agent_registry.increment_session_usage(session)
            # TODO: Read all download file contents. Need to fix by using chuncking, etc.
            results = await asyncio.gather(
                *map(
                    functools.partial(self._agent_registry.download_file, session),
                    files,
                ),
            )
            log.debug("file(s) inside container retrieved")
        except asyncio.CancelledError:
            raise
        except BackendAIError:
            raise
        except (ValueError, FileNotFoundError):
            raise InvalidAPIParameters("The file is not found.")
        except Exception as e:
            await self._error_monitor.capture_exception(context={"user": user_id})
            log.exception("DOWNLOAD_FILE: unexpected error!", e)
            raise InternalServerError

        with aiohttp.MultipartWriter("mixed") as mpwriter:
            headers = multidict.MultiDict({"Content-Encoding": "identity"})
            for tarbytes in results:
                mpwriter.append(tarbytes, headers)

            return DownloadFilesActionResult(
                session_data=session.to_dataclass(),
                result=mpwriter,  # type: ignore
            )

    async def execute_session(self, action: ExecuteSessionAction) -> ExecuteSessionActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        api_version = action.api_version

        resp = {}
        session = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )
        try:
            await self._agent_registry.increment_session_usage(session)

            run_id: Optional[str]
            mode: Optional[str]

            if api_version[0] == 1:
                run_id = action.params.run_id or secrets.token_hex(8)
                mode = "query"
                code = action.params.code
                opts = None
            elif api_version[0] >= 2:
                run_id = action.params.run_id
                mode = action.params.mode

                if mode is None:
                    raise InvalidSessionData("runId or mode is missing")

                if mode not in {"query", "batch", "complete", "continue", "input"}:
                    raise InvalidSessionData(f"mode has an invalid value: {mode}")
                if mode in {"continue", "input"} and run_id is None:
                    raise InvalidSessionData("continuation requires explicit run ID")
                code = action.params.code
                opts = action.params.options
            else:
                raise RuntimeError("should not reach here")
            # handle cases when some params are deliberately set to None
            if code is None:
                code = ""  # noqa
            if opts is None:
                opts = {}  # noqa
            if mode == "complete":
                # For legacy
                completion_resp = await self._agent_registry.get_completions(session, code, opts)
                resp["result"] = completion_resp.as_dict()
            else:
                run_id = cast(str, run_id)
                raw_result = await self._agent_registry.execute(
                    session,
                    api_version,
                    run_id,
                    mode,
                    code,
                    opts,
                    flush_timeout=2.0,
                )
                if raw_result is None:
                    # the kernel may have terminated from its side,
                    # or there was interruption of agents.
                    resp["result"] = {
                        "status": "finished",
                        "runId": run_id,
                        "exitCode": 130,
                        "options": {},
                        "files": [],
                        "console": [],
                    }
                    return ExecuteSessionActionResult(
                        result=resp, session_data=session.to_dataclass()
                    )
                # Keep internal/public API compatilibty
                result = {
                    "status": raw_result["status"],
                    "runId": raw_result["runId"],
                    "exitCode": raw_result.get("exitCode"),
                    "options": raw_result.get("options"),
                    "files": raw_result.get("files"),
                }
                if api_version[0] == 1:
                    result["stdout"] = raw_result.get("stdout")
                    result["stderr"] = raw_result.get("stderr")
                    result["media"] = raw_result.get("media")
                    result["html"] = raw_result.get("html")
                else:
                    result["console"] = raw_result.get("console")
                resp["result"] = result
        except AssertionError as e:
            log.warning("EXECUTE: invalid/missing parameters: {0!r}", e)
            raise InvalidAPIParameters(extra_msg=e.args[0])
        except BackendAIError:
            raise

        return ExecuteSessionActionResult(result=resp, session_data=session.to_dataclass())

    async def get_abusing_report(
        self, action: GetAbusingReportAction
    ) -> GetAbusingReportActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        session = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )
        kernel = session.main_kernel
        report = await self._agent_registry.get_abusing_report(kernel.id)
        return GetAbusingReportActionResult(
            abuse_report=report, session_data=session.to_dataclass()
        )

    async def get_commit_status(self, action: GetCommitStatusAction) -> GetCommitStatusActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key

        session = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )
        statuses = await self._agent_registry.get_commit_status([session.main_kernel.id])

        commit_info = CommitStatusInfo(
            status=statuses[session.main_kernel.id], kernel=str(session.main_kernel.id)
        )
        return GetCommitStatusActionResult(
            commit_info=commit_info, session_data=session.to_dataclass()
        )

    async def get_container_logs(
        self, action: GetContainerLogsAction
    ) -> GetContainerLogsActionResult:
        resp = {"result": {"logs": ""}}
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        kernel_id = action.kernel_id

        compute_session = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            allow_stale=True,
            kernel_loading_strategy=(
                KernelLoadingStrategy.MAIN_KERNEL_ONLY
                if kernel_id is None
                else KernelLoadingStrategy.ALL_KERNELS
            ),
        )

        if compute_session.status in DEAD_SESSION_STATUSES:
            if kernel_id is None:
                # Get logs from the main kernel
                kernel_id = compute_session.main_kernel.id
                kernel_log = compute_session.main_kernel.container_log
            else:
                # Get logs from the specific kernel
                kernel_row = compute_session.get_kernel_by_id(kernel_id)
                kernel_log = kernel_row.container_log
            if kernel_log is not None:
                # Get logs from database record
                log.debug("returning log from database record")
                resp["result"]["logs"] = kernel_log.decode("utf-8")
                return GetContainerLogsActionResult(
                    result=resp, session_data=compute_session.to_dataclass()
                )

        registry = self._agent_registry
        await registry.increment_session_usage(compute_session)
        resp["result"]["logs"] = await registry.get_logs_from_agent(
            session=compute_session, kernel_id=kernel_id
        )
        log.debug("returning log from agent")

        return GetContainerLogsActionResult(
            result=resp, session_data=compute_session.to_dataclass()
        )

    async def get_dependency_graph(
        self, action: GetDependencyGraphAction
    ) -> GetDependencyGraphActionResult:
        root_session_name = action.root_session_name
        owner_access_key = action.owner_access_key

        dependency_graph = await self._session_repository.find_dependency_sessions(
            root_session_name, owner_access_key
        )

        session_id = (
            dependency_graph.get("session_id") if isinstance(dependency_graph, dict) else None
        )
        if session_id:
            if isinstance(session_id, list) and session_id:
                session_id = session_id[0]
            session = await self._session_repository.get_session_by_id(str(session_id))
        else:
            session = None

        if session is None:
            raise SessionNotFound
        return GetDependencyGraphActionResult(
            result=dependency_graph, session_data=session.to_dataclass()
        )

    async def get_direct_access_info(
        self, action: GetDirectAccessInfoAction
    ) -> GetDirectAccessInfoActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key

        sess = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )
        resp = {}
        sess_type = cast(SessionTypes, sess.session_type)
        if sess_type in PRIVATE_SESSION_TYPES:
            public_host = sess.main_kernel.agent_row.public_host
            found_ports: dict[str, list[str]] = {}
            service_ports = cast(Optional[list[dict[str, Any]]], sess.main_kernel.service_ports)
            if service_ports is None:
                raise KernelNotReady(
                    f"Kernel of the session has no service ports yet (kernel: {sess.main_kernel.id}, kernel status: {sess.main_kernel.status.name})"
                )
            for sport in service_ports:
                if sport["name"] == "sshd":
                    found_ports["sshd"] = sport["host_ports"]
                elif sport["name"] == "sftpd":
                    found_ports["sftpd"] = sport["host_ports"]
            resp = {
                "kernel_role": sess_type.name,  # legacy
                "session_type": sess_type.name,
                "public_host": public_host,
                "sshd_ports": found_ports.get("sftpd") or found_ports["sshd"],
            }

        return GetDirectAccessInfoActionResult(result=resp, session_data=sess.to_dataclass())

    async def get_session_info(self, action: GetSessionInfoAction) -> GetSessionInfoActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key

        sess = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )
        await self._agent_registry.increment_session_usage(sess)

        age = datetime.now(tzutc()) - sess.created_at
        session_info = LegacySessionInfo(
            domain_name=sess.domain_name,
            group_id=sess.group_id,
            user_id=sess.user_uuid,
            lang=sess.main_kernel.image,  # legacy
            image=sess.main_kernel.image,
            architecture=sess.main_kernel.architecture,
            registry=sess.main_kernel.registry,
            tag=sess.tag,
            container_id=sess.main_kernel.container_id,
            occupied_slots=str(sess.main_kernel.occupied_slots),  # legacy
            occupying_slots=str(sess.occupying_slots),
            requested_slots=str(sess.requested_slots),
            occupied_shares=str(sess.main_kernel.occupied_shares),  # legacy
            environ=str(sess.environ),
            resource_opts=str(sess.resource_opts),
            status=sess.status.name,
            status_info=str(sess.status_info) if sess.status_info else None,
            status_data=sess.status_data,
            age_ms=int(age.total_seconds() * 1000),
            creation_time=sess.created_at,
            termination_time=sess.terminated_at,
            num_queries_executed=sess.num_queries,
            last_stat=sess.last_stat,
            idle_checks=await self._idle_checker_host.get_idle_check_report(sess.id),
        )

        # Resource limits collected from agent heartbeats were erased, as they were deprecated
        # TODO: factor out policy/image info as a common repository
        return GetSessionInfoActionResult(
            session_info=session_info, session_data=sess.to_dataclass()
        )

    async def get_status_history(
        self, action: GetStatusHistoryAction
    ) -> GetStatusHistoryActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key

        session_row = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.NONE,
        )
        result = session_row.status_history

        return GetStatusHistoryActionResult(status_history=result, session_id=session_row.id)

    async def interrupt(self, action: InterruptSessionAction) -> InterruptSessionActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key

        session = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )
        await self._agent_registry.increment_session_usage(session)
        await self._agent_registry.interrupt_session(session)

        return InterruptSessionActionResult(result=None, session_data=session.to_dataclass())

    async def list_files(self, action: ListFilesAction) -> ListFilesActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        user_id = action.user_id
        path = action.path

        session = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )

        resp: MutableMapping[str, Any] = {}
        try:
            await self._agent_registry.increment_session_usage(session)
            result = await self._agent_registry.list_files(session, path)
            resp.update(result)
            log.debug("container file list for {0} retrieved", path)
        except asyncio.CancelledError:
            raise
        except BackendAIError:
            raise
        except Exception as e:
            await self._error_monitor.capture_exception(context={"user": user_id})
            log.exception("LIST_FILES: unexpected error!", e)
            raise InternalServerError

        return ListFilesActionResult(result=result, session_data=session.to_dataclass())

    async def match_sessions(self, action: MatchSessionsAction) -> MatchSessionsActionResult:
        id_or_name_prefix = action.id_or_name_prefix
        owner_access_key = action.owner_access_key

        matches: list[dict[str, Any]] = []
        sessions = await self._session_repository.match_sessions(
            id_or_name_prefix,
            owner_access_key,
        )
        if sessions:
            matches.extend(
                {
                    "id": str(item.id),
                    "name": item.name,
                    "status": item.status.name,
                }
                for item in sessions
            )
        return MatchSessionsActionResult(result=matches)

    async def rename_session(self, action: RenameSessionAction) -> RenameSessionActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        new_name = action.new_name

        try:
            compute_session = await self._session_repository.update_session_name(
                session_name, new_name, owner_access_key
            )
            if compute_session.status != SessionStatus.RUNNING:
                raise InvalidAPIParameters("Can't change name of not running session")
        except ValueError as e:
            if "already exists" in str(e):
                raise InvalidAPIParameters(str(e))
            raise

        return RenameSessionActionResult(session_data=compute_session.to_dataclass())

    async def restart_session(self, action: RestartSessionAction) -> RestartSessionActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key

        session = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.ALL_KERNELS,
        )
        await self._agent_registry.increment_session_usage(session)
        await self._agent_registry.restart_session(session)
        return RestartSessionActionResult(result=None, session_data=session.to_dataclass())

    async def shutdown_service(self, action: ShutdownServiceAction) -> ShutdownServiceActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        service_name = action.service_name

        session = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )
        await self._agent_registry.shutdown_service(session, service_name)
        return ShutdownServiceActionResult(result=None, session_data=session.to_dataclass())

    async def start_service(self, action: StartServiceAction) -> StartServiceActionResult:
        session_name = action.session_name
        access_key = action.access_key
        service = action.service
        port = action.port

        arguments = action.arguments
        envs = action.envs
        login_session_token = action.login_session_token

        session = await asyncio.shield(
            self._database_ptask_group.create_task(
                self._session_repository.get_session_with_routing_minimal(
                    session_name,
                    access_key,
                )
            )
        )

        wsproxy_addr = await self._session_repository.get_scaling_group_wsproxy_addr(
            session.scaling_group_name
        )
        if not wsproxy_addr:
            raise ServiceUnavailable("No coordinator configured for this resource group")
        wsproxy_status = await query_wsproxy_status(wsproxy_addr)
        if advertise_addr := wsproxy_status.get("advertise_address"):
            wsproxy_advertise_addr = advertise_addr
        else:
            wsproxy_advertise_addr = wsproxy_addr

        if session.main_kernel.kernel_host is None:
            kernel_host = urlparse(session.main_kernel.agent_addr).hostname
        else:
            kernel_host = session.main_kernel.kernel_host
        for sport in session.main_kernel.service_ports:
            if sport["name"] == service:
                if sport["is_inference"]:
                    raise InvalidAPIParameters(
                        f"{service} is an inference app. Starting inference apps can only be done by"
                        " starting an inference service."
                    )
                if port:
                    # using one of the primary/secondary ports of the app
                    try:
                        hport_idx = sport["container_ports"].index(port)
                    except ValueError:
                        raise InvalidAPIParameters(
                            f"Service {service} does not open the port number {port}."
                        )
                    host_port = sport["host_ports"][hport_idx]
                else:
                    # using the default (primary) port of the app
                    if "host_ports" not in sport:
                        host_port = sport["host_port"]  # legacy kernels
                    else:
                        host_port = sport["host_ports"][0]
                break
        else:
            raise AppNotFound(f"{session_name}:{service}")

        await asyncio.shield(
            self._database_ptask_group.create_task(
                self._agent_registry.increment_session_usage(session),
            )
        )

        opts: MutableMapping[str, Union[None, str, list[str]]] = {}
        if arguments is not None:
            opts["arguments"] = load_json(arguments)
        if envs is not None:
            opts["envs"] = load_json(envs)

        result = await asyncio.shield(
            self._rpc_ptask_group.create_task(
                self._agent_registry.start_service(session, service, opts),
            ),
        )
        if result["status"] == "failed":
            raise InternalServerError(
                "Failed to launch the app service", extra_data=result["error"]
            )

        body = {
            "login_session_token": login_session_token,
            "kernel_host": kernel_host,
            "kernel_port": host_port,
            "session": {
                "id": str(session.id),
                "user_uuid": str(session.user_uuid),
                "group_id": str(session.group_id),
                "access_key": session.access_key,
                "domain_name": session.domain_name,
            },
        }

        async with aiohttp.ClientSession() as req:
            async with req.post(
                f"{wsproxy_addr}/v2/conf",
                json=body,
            ) as resp:
                token_json = await resp.json()

                return StartServiceActionResult(
                    result=None,
                    session_data=session.to_dataclass(),
                    token=token_json["token"],
                    wsproxy_addr=wsproxy_advertise_addr,
                )

    async def upload_files(self, action: UploadFilesAction) -> UploadFilesActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        reader = action.reader

        session = await self._session_repository.get_session_validated(
            session_name,
            owner_access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )

        await self._agent_registry.increment_session_usage(session)
        file_count = 0

        async with aiotools.TaskScope() as ts:
            async for file in aiotools.aiter(reader.next, None):
                if file_count == 20:
                    raise InvalidAPIParameters("Too many files")
                if file is None:
                    break
                if not isinstance(file, BodyPartReader):
                    raise InvalidAPIParameters("Nested multipart upload is not supported")
                file_name = file.filename or f"upload-{secrets.token_hex(12)}"
                file_count += 1
                # This API handles only small files, so let's read it at once.
                chunks = []
                recv_size = 0
                while True:
                    chunk = await file.read_chunk(size=1048576)
                    if not chunk:
                        break
                    chunk_size = len(chunk)
                    if recv_size + chunk_size >= 1048576:
                        raise InvalidAPIParameters("Too large file")
                    chunks.append(chunk)
                    recv_size += chunk_size
                data = file.decode(b"".join(chunks))
                log.debug("received file: {0} ({1:,} bytes)", file_name, recv_size)
                ts.create_task(self._agent_registry.upload_file(session, file_name, data))

        return UploadFilesActionResult(result=None, session_data=session.to_dataclass())

    async def modify_session(self, action: ModifySessionAction) -> ModifySessionActionResult:
        session_id = action.session_id
        props = action.modifier
        session_name = action.modifier.name.optional_value()

        session_row = await self._session_repository.modify_session(
            str(session_id), props.fields_to_update(), session_name
        )
        if session_row is None:
            raise ValueError(f"Session not found (id:{session_id})")
        session_owner_data = await self._session_repository.get_session_owner(str(session_id))

        return ModifySessionActionResult(
            session_data=session_row.to_dataclass(owner=session_owner_data)
        )

    async def check_and_transit_status(
        self, action: CheckAndTransitStatusAction
    ) -> CheckAndTransitStatusActionResult:
        user_id = action.user_id
        user_role = action.user_role
        session_id = action.session_id

        if user_role in (UserRole.ADMIN, UserRole.SUPERADMIN):
            session_row = (
                await self._admin_session_repository.get_session_to_determine_status_force(
                    session_id
                )
            )
        else:
            session_row = await self._session_repository.get_session_to_determine_status(session_id)
            if session_row.user_uuid != user_id:
                log.warning(
                    f"You are not allowed to transit others's sessions status, skip (s:{session_id})"
                )
                return CheckAndTransitStatusActionResult(
                    result={}, session_data=session_row.to_dataclass()
                )

        now = datetime.now(tzutc())
        session_rows = await self._agent_registry.session_lifecycle_mgr.transit_session_status(
            [session_id], now
        )
        await self._agent_registry.session_lifecycle_mgr.deregister_status_updatable_session([
            row.id for row, is_transited in session_rows if is_transited
        ])
        session_owner_data = await self._session_repository.get_session_owner(session_id)

        result = {row.id: row.status.name for row, _ in session_rows}
        return CheckAndTransitStatusActionResult(
            result=result, session_data=session_row.to_dataclass(owner=session_owner_data)
        )

    async def check_and_transit_status_multi(
        self, action: CheckAndTransitStatusBatchAction
    ) -> CheckAndTransitStatusBatchActionResult:
        user_id = action.user_id
        user_role = action.user_role
        session_ids = action.session_ids
        accessible_session_ids: list[SessionId] = []

        for sid in session_ids:
            if user_role in (UserRole.ADMIN, UserRole.SUPERADMIN):
                accessible_session_ids.append(sid)
            else:
                try:
                    session_row = await self._session_repository.get_session_to_determine_status(
                        sid
                    )
                    if session_row.user_uuid == user_id:
                        accessible_session_ids.append(sid)
                    else:
                        log.warning(
                            f"You are not allowed to transit others's sessions status, skip (s:{sid})"
                        )
                except Exception:
                    log.warning(f"Session not found or access denied, skip (s:{sid})")

        now = datetime.now(tzutc())
        if accessible_session_ids:
            session_rows = await self._agent_registry.session_lifecycle_mgr.transit_session_status(
                accessible_session_ids, now
            )
            await self._agent_registry.session_lifecycle_mgr.deregister_status_updatable_session([
                row.id for row, is_transited in session_rows if is_transited
            ])
            result = {row.id: row.status.name for row, _ in session_rows}
        else:
            result = {}

        return CheckAndTransitStatusBatchActionResult(session_status_map=result)



from __future__ import annotations

import enum
from collections.abc import Iterable, Mapping
from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, Self, override
from uuid import UUID

from sqlalchemy.engine import Row

from ai.backend.common.types import AccessKey, CIStrEnum
from ai.backend.manager.data.permission.id import ScopeId
from ai.backend.manager.data.permission.types import (
    EntityType,
    OperationType,
    ScopeType,
)
from ai.backend.manager.errors.resource import DataTransformationFailed
from ai.backend.manager.types import Creator

from ..keypair.types import KeyPairData

if TYPE_CHECKING:
    from ai.backend.manager.models.hasher.types import PasswordInfo


class UserStatus(enum.StrEnum):
    """
    User account status.
    """

    ACTIVE = "active"
    INACTIVE = "inactive"
    DELETED = "deleted"
    BEFORE_VERIFICATION = "before-verification"

    @override
    @classmethod
    def _missing_(cls, value: Any) -> Optional[UserStatus]:
        if not isinstance(value, str):
            raise DataTransformationFailed(
                f"UserStatus value must be a string, got {type(value).__name__}"
            )
        match value.upper():
            case "ACTIVE":
                return cls.ACTIVE
            case "INACTIVE":
                return cls.INACTIVE
            case "DELETED":
                return cls.DELETED
            case "BEFORE-VERIFICATION" | "BEFORE_VERIFICATION":
                return cls.BEFORE_VERIFICATION
        return None


class UserRole(CIStrEnum):
    """
    User's role.
    """

    SUPERADMIN = "superadmin"
    ADMIN = "admin"
    USER = "user"
    MONITOR = "monitor"


@dataclass
class UserCreator(Creator):
    email: str
    username: str
    password: PasswordInfo  # Only accept PasswordInfo
    need_password_change: bool
    domain_name: str
    full_name: Optional[str] = None
    description: Optional[str] = None
    is_active: Optional[bool] = None
    status: Optional[UserStatus] = None
    role: Optional[str] = None
    allowed_client_ip: Optional[list[str]] = None
    totp_activated: Optional[bool] = None
    resource_policy: Optional[str] = None
    sudo_session_enabled: Optional[bool] = None
    container_uid: Optional[int] = None
    container_main_gid: Optional[int] = None
    container_gids: Optional[list[int]] = None

    def fields_to_store(self) -> dict[str, Any]:
        status = UserStatus.ACTIVE  # TODO: Need to be set in action explicitly not in service (integrate is_active and status)
        if self.status is None and self.is_active is not None:
            status = UserStatus.ACTIVE if self.is_active else UserStatus.INACTIVE
        if self.status is not None:
            status = self.status
        user_data = {
            "email": self.email,
            "username": self.username,
            "password": self.password,
            "need_password_change": self.need_password_change,
            "domain_name": self.domain_name,
            "full_name": self.full_name,
            "description": self.description,
            "status": status,
            "role": self.role,
            "allowed_client_ip": self.allowed_client_ip,
            "totp_activated": self.totp_activated,
            "resource_policy": self.resource_policy,
            "sudo_session_enabled": self.sudo_session_enabled,
            "container_uid": self.container_uid,
            "container_main_gid": self.container_main_gid,
            "container_gids": self.container_gids,
        }
        if self.container_uid is not None:
            user_data["container_uid"] = self.container_uid
        if self.container_main_gid is not None:
            user_data["container_main_gid"] = self.container_main_gid
        if self.container_gids is not None:
            user_data["container_gids"] = self.container_gids
        return user_data


@dataclass
class UserInfoContext:
    uuid: UUID
    email: str
    main_access_key: AccessKey


@dataclass
class UserData:
    id: UUID = field(compare=False)
    uuid: UUID = field(compare=False)  # legacy
    username: str
    email: str
    need_password_change: bool
    full_name: Optional[str]
    description: Optional[str]
    is_active: bool  # legacy
    status: str
    status_info: Optional[str]
    created_at: datetime = field(compare=False)
    modified_at: datetime = field(compare=False)
    domain_name: str
    role: UserRole
    resource_policy: str
    allowed_client_ip: Optional[list[str]]
    totp_activated: bool
    totp_activated_at: Optional[datetime] = field(compare=False)
    sudo_session_enabled: bool
    main_access_key: Optional[str] = field(compare=False)
    container_uid: Optional[int] = field(compare=False)
    container_main_gid: Optional[int] = field(compare=False)
    container_gids: Optional[list[int]] = field(compare=False)

    def scope_id(self) -> ScopeId:
        return ScopeId(
            scope_type=ScopeType.USER,
            scope_id=str(self.id),
        )

    def role_name(self) -> str:
        return f"user-{str(self.id)[:8]}"

    def entity_operations(self) -> Mapping[EntityType, Iterable[OperationType]]:
        resource_entity_permissions = {
            entity: OperationType.owner_operations()
            for entity in EntityType.owner_accessible_entity_types_in_user()
        }
        user_permissions = OperationType.owner_operations() - {OperationType.CREATE}
        return {EntityType.USER: user_permissions, **resource_entity_permissions}

    @classmethod
    def from_row(cls, row: Row) -> Self:
        """
        Deprecated: Use `UserRow.to_data()` method instead.
        """
        return cls(
            id=row.uuid,
            uuid=row.uuid,
            username=row.username,
            email=row.email,
            need_password_change=row.need_password_change,
            full_name=row.full_name,
            description=row.description,
            is_active=True if row.status == UserStatus.ACTIVE else False,
            status=row.status,
            status_info=row.status_info,
            created_at=row.created_at,
            modified_at=row.modified_at,
            domain_name=row.domain_name,
            role=row.role,
            resource_policy=row.resource_policy,
            allowed_client_ip=row.allowed_client_ip,
            totp_activated=row.totp_activated,
            totp_activated_at=row.totp_activated_at,
            sudo_session_enabled=row.sudo_session_enabled,
            main_access_key=row.main_access_key,
            container_uid=row.container_uid,
            container_main_gid=row.container_main_gid,
            container_gids=row.container_gids,
        )


@dataclass
class UserCreateResultData:
    user: UserData
    keypair: KeyPairData



from __future__ import annotations

import uuid
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Optional, override

from ai.backend.manager.types import OptionalState, PartialModifier, TriState

from .id import ObjectId, ScopeId
from .object_permission import (
    ObjectPermissionCreateInputBeforeRoleCreation,
    ObjectPermissionData,
)
from .permission_group import (
    PermissionGroupCreatorBeforeRoleCreation,
    PermissionGroupExtendedData,
)
from .status import RoleStatus
from .types import EntityType, OperationType, RoleSource


@dataclass(frozen=True)
class RoleCreateInput:
    name: str
    source: RoleSource = RoleSource.CUSTOM
    status: RoleStatus = RoleStatus.ACTIVE
    description: Optional[str] = None

    permission_groups: list[PermissionGroupCreatorBeforeRoleCreation] = field(default_factory=list)
    object_permissions: list[ObjectPermissionCreateInputBeforeRoleCreation] = field(
        default_factory=list
    )


@dataclass(frozen=True)
class RoleUpdateInput(PartialModifier):
    id: uuid.UUID
    name: OptionalState[str]
    source: OptionalState[RoleSource]
    status: OptionalState[RoleStatus]
    description: TriState[str]

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.source.update_dict(to_update, "source")
        self.status.update_dict(to_update, "status")
        self.description.update_dict(to_update, "description")
        return to_update


@dataclass(frozen=True)
class RoleDeleteInput:
    id: uuid.UUID
    _status: RoleStatus = RoleStatus.DELETED


@dataclass(frozen=True)
class RoleData:
    id: uuid.UUID
    name: str
    source: RoleSource
    status: RoleStatus
    created_at: datetime
    updated_at: Optional[datetime]
    deleted_at: Optional[datetime]
    description: Optional[str] = None


@dataclass(frozen=True)
class RoleDataWithPermissions:
    id: uuid.UUID
    name: str
    source: RoleSource
    status: RoleStatus

    permission_groups: list[PermissionGroupExtendedData]
    object_permissions: list[ObjectPermissionData]

    created_at: datetime
    updated_at: Optional[datetime]
    deleted_at: Optional[datetime]
    description: Optional[str] = None


@dataclass(frozen=True)
class ScopePermissionCheckInput:
    user_id: uuid.UUID
    target_entity_type: EntityType
    target_scope_id: ScopeId
    operation: OperationType


@dataclass(frozen=True)
class SingleEntityPermissionCheckInput:
    user_id: uuid.UUID
    target_object_id: ObjectId
    operation: OperationType


@dataclass(frozen=True)
class BatchEntityPermissionCheckInput:
    user_id: uuid.UUID
    target_object_ids: list[ObjectId]
    operation: OperationType


@dataclass(frozen=True)
class UserRoleAssignmentInput:
    """
    Input to create a new user-role association.
    """

    user_id: uuid.UUID
    role_id: uuid.UUID
    granted_by: Optional[uuid.UUID] = None


@dataclass(frozen=True)
class UserRoleAssignmentData:
    user_id: uuid.UUID
    role_id: uuid.UUID
    granted_by: Optional[uuid.UUID] = None



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.artifact_registries.types import ArtifactRegistryModifierMeta
from ai.backend.manager.data.huggingface_registry.modifier import HuggingFaceRegistryModifier
from ai.backend.manager.data.huggingface_registry.types import HuggingFaceRegistryData
from ai.backend.manager.services.artifact_registry.actions.base import ArtifactRegistryAction


@dataclass
class UpdateHuggingFaceRegistryAction(ArtifactRegistryAction):
    id: uuid.UUID
    modifier: HuggingFaceRegistryModifier
    meta: ArtifactRegistryModifierMeta

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "update_huggingface_registry"


@dataclass
class UpdateHuggingFaceRegistryActionResult(BaseActionResult):
    result: HuggingFaceRegistryData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.result.id)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.notification import (
    NotificationChannelData,
    NotificationChannelModifier,
)

from .base import NotificationAction


@dataclass
class UpdateChannelAction(NotificationAction):
    """Action to update a notification channel."""

    channel_id: UUID
    modifier: NotificationChannelModifier

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "update_channel"

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.channel_id)


@dataclass
class UpdateChannelActionResult(BaseActionResult):
    """Result of updating a notification channel."""

    channel_data: NotificationChannelData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.channel_data.id)



from dataclasses import dataclass

from ai.backend.common.exception import RuntimeVariantNotSupportedError
from ai.backend.common.types import RuntimeVariant
from ai.backend.manager.repositories.deployment import DeploymentRepository
from ai.backend.manager.sokovan.deployment.definition_generator.base import ModelDefinitionGenerator
from ai.backend.manager.sokovan.deployment.definition_generator.cmd import (
    CMDModelDefinitionGenerator,
)
from ai.backend.manager.sokovan.deployment.definition_generator.custom import (
    CustomModelDefinitionGenerator,
)
from ai.backend.manager.sokovan.deployment.definition_generator.huggingface_tgi import (
    HuggingFaceTGIModelDefinitionGenerator,
)
from ai.backend.manager.sokovan.deployment.definition_generator.modular_max import (
    ModularMAXModelDefinitionGenerator,
)
from ai.backend.manager.sokovan.deployment.definition_generator.nim import (
    NIMModelDefinitionGenerator,
)
from ai.backend.manager.sokovan.deployment.definition_generator.sglang import (
    SGLangModelDefinitionGenerator,
)
from ai.backend.manager.sokovan.deployment.definition_generator.vllm import (
    VLLMModelDefinitionGenerator,
)


@dataclass
class RegistryArgs:
    deployment_repository: DeploymentRepository


class ModelDefinitionGeneratorRegistry:
    _generators: dict[RuntimeVariant, ModelDefinitionGenerator]

    def __init__(self, args: RegistryArgs) -> None:
        self._generators: dict[RuntimeVariant, ModelDefinitionGenerator] = {}
        self._generators[RuntimeVariant.CUSTOM] = CustomModelDefinitionGenerator(
            args.deployment_repository
        )
        self._generators[RuntimeVariant.VLLM] = VLLMModelDefinitionGenerator()
        self._generators[RuntimeVariant.HUGGINGFACE_TGI] = HuggingFaceTGIModelDefinitionGenerator()
        self._generators[RuntimeVariant.NIM] = NIMModelDefinitionGenerator()
        self._generators[RuntimeVariant.SGLANG] = SGLangModelDefinitionGenerator()
        self._generators[RuntimeVariant.MODULAR_MAX] = ModularMAXModelDefinitionGenerator()
        self._generators[RuntimeVariant.CMD] = CMDModelDefinitionGenerator()

    def get(self, runtime_variant: RuntimeVariant) -> ModelDefinitionGenerator:
        generator = self._generators.get(runtime_variant)
        if generator is None:
            raise RuntimeVariantNotSupportedError(runtime_variant)
        return generator



"""
Result type for scheduling operations.
"""

from dataclasses import dataclass, field

from ai.backend.common.types import AccessKey, SessionId


@dataclass
class ScheduledSessionData:
    """Data for a scheduled session."""

    session_id: SessionId
    creation_id: str
    access_key: AccessKey
    reason: str


@dataclass
class ScheduleResult:
    """Result of a scheduling operation."""

    # List of scheduled session data
    scheduled_sessions: list[ScheduledSessionData] = field(default_factory=list)

    def needs_post_processing(self) -> bool:
        """Check if post-processing is needed based on the result."""
        return len(self.scheduled_sessions) > 0

    def success_count(self) -> int:
        """Get the count of successfully scheduled sessions."""
        return len(self.scheduled_sessions)



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.common.resource.types import TotalResourceData
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.agent.actions.base import AgentAction


@dataclass
class GetTotalResourcesAction(AgentAction):
    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "get_total_resources"


@dataclass
class GetTotalResourcesActionResult(BaseActionResult):
    total_resources: TotalResourceData

    @override
    def entity_id(self) -> Optional[str]:
        return None



from __future__ import annotations

from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, AsyncIterator

if TYPE_CHECKING:
    from .api.context import RootContext


@asynccontextmanager
async def gql_adapters_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    """
    Initialize GraphQL adapters.

    These adapters are created once at server startup and reused across all GraphQL requests.
    """
    from .api.gql.adapter import BaseGQLAdapter

    root_ctx.gql_adapter = BaseGQLAdapter()

    yield
